#!/usr/bin/env python3
import glob
import logging
import os
import socket
import statistics
import warnings
import zoneinfo
from dataclasses import dataclass
from datetime import date, datetime, time, timedelta
from multiprocessing import Pool
from pathlib import Path
from typing import Dict, List, Tuple, Union

import eflips.model
import fire  # type: ignore
import psycopg2
from eflips.model import ConsistencyWarning
from geoalchemy2 import WKBElement
from geoalchemy2.functions import ST_Distance
from geoalchemy2.shape import to_shape
from lxml import etree
from sqlalchemy import create_engine, func
from sqlalchemy.orm import Session
from tqdm.auto import tqdm
from xsdata.formats.dataclass.parsers import XmlParser

import eflips.ingest.util
from eflips.ingest.legacy.xmldata import (
    Linienfahrplan,
    NetzpunktNetzpunkttyp,
)
from eflips.ingest.util import soldner_to_pointz


def load_and_validate_xml(filename: Path) -> Linienfahrplan:
    """
    Loads the xml string into a Linienfahrplan object and validates it against the schema
    - also removes the 'ns2' strings from the first and last lines

    :param filename: the filename to load

    :return: a Linienfahrplan object of there is data, None otherwise. raises a ValueError if the data is there but
             not valid
    """
    xml_string = filename.read_text()
    logger = logging.getLogger(__name__)

    if "Keine gültige Linie." in xml_string:
        raise ValueError(f"File {filename} is not a valid line.")
    elif "Keine Umläufe vorhanden." in xml_string:
        raise ValueError(f"File {filename} does not contain any rotations.")

    # We want to remove the occurrences of 'ns2: and ':ns2' in the first and last line. Only then does our schema
    # and the 'xmldata' package work. For some reason, with the ns2 it generates two differnet python files, and
    # then it doesn't work.
    xml_string = xml_string.replace("ns2:", "")
    xml_string = xml_string.replace(":ns2", "")

    xsd_path = Path(__file__).parent.parent.parent.parent / "data" / "bvg_xml.xsd"
    xmlschema_doc = etree.parse(xsd_path)
    xmlschema = etree.XMLSchema(xmlschema_doc)

    xml_doc = etree.fromstring(xml_string)

    if not xmlschema.validate(xml_doc):
        raise ValueError(f"XML file {filename} is not valid.")

    parser = XmlParser()

    data: Linienfahrplan = parser.from_string(xml_string, Linienfahrplan)

    return data


def add_or_ret_station(scenario_id: int, id: int, name: str, name_short: str, session: Session) -> eflips.model.Station:
    """
    For a given station ID, adds a station to the database if it does not exist yet, and returns the station object.

    :param scenario_id: The scenario ID to use
    :param id: The station ID to use
    :param name: The long name of the station
    :param name_short: The short name of the station
    :param session: An open database session
    :return: A station object, connected to the database it will not have a geometry yet
    """

    station = (
        session.query(eflips.model.Station)
        .filter(eflips.model.Station.scenario_id == scenario_id)
        .filter(eflips.model.Station.id == id)
        .one_or_none()
    )
    if station is None:
        station = eflips.model.Station(
            scenario_id=scenario_id,
            id=id,
            name=name,
            name_short=name_short,
            is_electrified=False,
            geom="SRID=4326;POINTZ(0 0 0)",  # Will be set later
        )
        session.add(station)
    return station


def create_stations(linienfahrplan: Linienfahrplan, scenario_id: int, session: Session) -> None:
    """
    First method to be used when importing a set of xml files. It takes the parsed xml data and creates the stations
    from the 'Linienfahrplan/StreckennetzDaten/Haltestellenbereiche/Haltestellenbereich' entries.

    :param linienfahrplan: A parsed Linienfahrplan object
    :param scenario_id: The scenario ID to use
    :param session: An open database session
    :return: Nothing - the stations are added to the database
    """
    logger = logging.getLogger(__name__)

    for haltestellenbereich in linienfahrplan.streckennetz_daten.haltestellenbereiche.haltestellenbereich:
        id_no = haltestellenbereich.nummer
        short_name = haltestellenbereich.kurzname
        long_name = haltestellenbereich.fahrplanbuchname

        add_or_ret_station(scenario_id, id_no, long_name, short_name, session)


def add_or_ret_line(scenario_id: int, name: str, session: Session) -> eflips.model.Line:
    line = (
        session.query(eflips.model.Line)
        .filter(eflips.model.Scenario.id == scenario_id)
        .filter(eflips.model.Line.name == name)
        .one_or_none()
    )
    if line is None:
        line = eflips.model.Line(scenario_id=scenario_id, name=name)
        session.add(line)
    return line


def add_or_ret_station_for_grid_point(
    scenario_id: int,
    gridpoint_id: int,
    grid_points: Dict[int, Linienfahrplan.StreckennetzDaten.Netzpunkte.Netzpunkt],
    session: Session,
) -> eflips.model.Station:
    """
    Sets up the station object for a given grid point. If it is if the `Netzpunkttyp` "Hst", returns the station.
    Otherwise, check if a station already exists for the grid point (by short name). If not, create a new station
    """

    logger = logging.getLogger(__name__)

    grid_point = grid_points[gridpoint_id]
    if grid_point.netzpunkttyp == NetzpunktNetzpunkttyp.HST:
        station = (
            session.query(eflips.model.Station)
            .filter(eflips.model.Station.scenario_id == scenario_id)
            .filter(eflips.model.Station.id == grid_point.haltestellenbereich)
            .one_or_none()
        )
        if station is None:
            # The station should have already been created in the previous step
            raise ValueError(f"Station for grid point {gridpoint_id} not found, even though it is of type 'Hst'")
    elif grid_point.netzpunkttyp == NetzpunktNetzpunkttyp.BPUNKT:
        # Assumption: A "Betriebspunkt" basically belongs to the station for our purposes
        # We query the station by the four-character short name
        short_name = grid_point.kurzname[0:4]
        if short_name.endswith("0") or short_name.endswith("1"):
            # The 0 or 1 seems to be added to the short name for bpunkts corresponding to stations which actually have a
            # three-character short name. if we shorten the short name to three characters, we can find the station
            short_name = short_name[0:3]
        station = (
            session.query(eflips.model.Station)
            .filter(eflips.model.Station.scenario_id == scenario_id)
            .filter(eflips.model.Station.name_short == short_name)
            .one_or_none()
        )
        if station is None:
            logger.info(
                f"Station for grid point {gridpoint_id} not found, even though it is of type 'BPUNKT' and should have a station"
            )
            # Here, we can actually calculate the coordinates already
            geom = soldner_to_pointz(grid_point.xkoordinate, grid_point.ykoordinate)

            station = eflips.model.Station(
                scenario_id=scenario_id,
                name=f"BPUNKT {grid_point.langname}",
                name_short=short_name,
                is_electrified=False,
                geom=geom,
            )

            # raise ValueError(
            #    f"Station for grid point {gridpoint_id} not found, even though it is of type 'BPUNKT' and should have a station"
            # )
    elif grid_point.netzpunkttyp == NetzpunktNetzpunkttyp.EPKT or grid_point.netzpunkttyp == NetzpunktNetzpunkttyp.APKT:
        # We want to merge the Einsetzpunkt and Aussetzpunkt stations into one station
        # Make sure the short name ends in "E" or "A", then remove that character
        short_name = grid_point.kurzname
        if short_name[-1] != "E" and short_name[-1] != "A":
            raise ValueError(
                f"Station for grid point {gridpoint_id} not found, even though it is of type 'EPKT' or 'APKT' and should have a station"
            )
        short_name = short_name[:-2]

        # Make sure the long name ends in "Einsetzen" or "Aussetzen", then remove that word
        long_name = grid_point.langname
        if long_name[-9:] != "Einsetzen" and long_name[-9:] != "Aussetzen":
            raise ValueError(
                f"Station for grid point {gridpoint_id} not found, even though it is of type 'EPKT' or 'APKT' and should have a station"
            )
        long_name = long_name[:-10]

        station = (
            session.query(eflips.model.Station)
            .filter(eflips.model.Station.scenario_id == scenario_id)
            .filter(eflips.model.Station.name_short == short_name)
            .filter(eflips.model.Station.name == long_name)
            .one_or_none()
        )
        if station is None:
            # Here, we can actually calculate the coordinates already
            geom = soldner_to_pointz(grid_point.xkoordinate, grid_point.ykoordinate)

            station = eflips.model.Station(
                scenario_id=scenario_id,
                id=gridpoint_id,
                name=long_name,
                name_short=short_name,
                is_electrified=False,
                geom=geom,
            )
            session.add(station)
    else:
        raise ValueError(f"Grid point {gridpoint_id} is of type {grid_point.netzpunkttyp}, which is not supported")

    return station


def add_or_ret_vehicle_type(scenario_id: int, fahrzeugtyp: str, session: Session) -> eflips.model.VehicleType:
    vehicle_type = (
        session.query(eflips.model.VehicleType)
        .filter(eflips.model.VehicleType.scenario_id == scenario_id)
        .filter(eflips.model.VehicleType.name_short == fahrzeugtyp)
        .one_or_none()
    )
    if vehicle_type is None:
        vehicle_type = eflips.model.VehicleType(
            scenario_id=scenario_id,
            name=f"Auto-Generated by XML Importer for {fahrzeugtyp}",
            name_short=fahrzeugtyp,
            battery_capacity=1000,
            charging_curve=[[0, 150], [1, 150]],
            opportunity_charging_capable=True,
        )
        session.add(vehicle_type)
    return vehicle_type


def setup_working_dictionaries(
    schedule: Linienfahrplan,
) -> Tuple[
    Dict[int, Linienfahrplan.StreckennetzDaten.Netzpunkte.Netzpunkt],
    Dict[int, Linienfahrplan.StreckennetzDaten.Strecken.Strecke],
    Dict[int, Linienfahrplan.LinienDaten.Linie.RoutenDaten.Route],
    Dict[int, int],
]:
    """
    This creates a set of dictionaries that are used to assemble the routes and trips.

    - The first entry is a dict of the "Netzpunkte", by their ID.
    - The second entry is a dict of the "Strecken", by their ID.
    - The third entry is a dict of the "RoutenDaten", by their "lfd_nr"
    - The fourth dictionary is the "Route.lfd_nr" for each "Routenvariante.lfd_nr"

    :param schedule: The parsed Linienfahrplan object
    :return: A tuple of the four dictionaries
    """
    line_name = schedule.linien_daten.linie.kurzname

    # Create a dict from the "Netzpunkzte", to be used in reassembling the routes later on.
    grid_points: Dict[int, Linienfahrplan.StreckennetzDaten.Netzpunkte.Netzpunkt] = {}
    for netzpunkt in schedule.streckennetz_daten.netzpunkte.netzpunkt:
        grid_points[netzpunkt.nummer] = netzpunkt

    # Create a similar dict for the "Strecken", whoch will be the segments we use to assemble the route shape
    segments: Dict[int, Linienfahrplan.StreckennetzDaten.Strecken.Strecke] = {}
    for strecke in schedule.streckennetz_daten.strecken.strecke:
        segments[strecke.id] = strecke

    # Create a similar dict for the route data, which we will use to assemble the trips
    route_datas: Dict[int, Linienfahrplan.LinienDaten.Linie.RoutenDaten.Route] = {}
    for route in schedule.linien_daten.linie.routen_daten.route:
        route_datas[route.lfd_nr] = route

    # Create a list of the route lfd nr for all the variants
    route_lfd_nrs: Dict[int, int] = {}
    for route_variant in schedule.linien_daten.linie.routenvarianten.routenvariante:
        route_lfd_nrs[route_variant.lfd_nr] = route_variant.lfd_nr_route

    return grid_points, segments, route_datas, route_lfd_nrs


@dataclass
class TimeProfile:
    @dataclass
    class TimeProfilePoint:
        station: eflips.model.Station
        arrival_offset_from_start: timedelta
        dwell_duration: timedelta

        def __eq__(self: "TimeProfile.TimeProfilePoint", other: object) -> bool:
            if not isinstance(other, TimeProfile.TimeProfilePoint):
                return False
            if self.station != other.station:
                return False
            if self.arrival_offset_from_start != other.arrival_offset_from_start:
                return False
            if self.dwell_duration != other.dwell_duration:
                return False
            return True

    route: eflips.model.Route
    start_offset_from_midnight: timedelta
    time_profile_points: List[TimeProfilePoint]

    def __eq__(self: "TimeProfile", other: object) -> bool:
        if not isinstance(other, TimeProfile):
            return False
        if self.route != other.route:
            return False
        if self.start_offset_from_midnight != other.start_offset_from_midnight:
            return False
        if len(self.time_profile_points) != len(other.time_profile_points):
            return False
        for i in range(len(self.time_profile_points)):
            if self.time_profile_points[i] != other.time_profile_points[i]:
                return False
        return True

    def to_trip(
        self,
        rotation: eflips.model.Rotation,
        the_date: date,
        timezone: zoneinfo.ZoneInfo = zoneinfo.ZoneInfo("Europe/Berlin"),
    ) -> eflips.model.Trip:
        """
        Create a trip on a given date from this time profile

        :param rotation: The rotation to create the trip for
        :param the_date: The date to use
        :param timezone: The timezone in which the offsets from midnight are given. Defaults to Europe/Berlin
        :return: a trip object, which is not yet added to the database
        """
        logger = logging.getLogger(__name__)

        local_midnight = datetime.combine(the_date, time(0, 0, 0, 0), tzinfo=timezone)

        # Find the trip type fromt the route name:
        # - Einsetzfahrt
        # - Aussetzfahrt
        if "Einsetzfahrt" in self.route.name or "Aussetzfahrt" in self.route.name:
            trip_type = eflips.model.TripType.EMPTY
        else:
            trip_type = eflips.model.TripType.PASSENGER

        # Create the trip object
        trip = eflips.model.Trip(
            scenario_id=self.route.scenario_id,
            route=self.route,
            rotation=rotation,
            departure_time=local_midnight
            + self.start_offset_from_midnight
            + self.time_profile_points[0].arrival_offset_from_start,
            arrival_time=local_midnight
            + self.start_offset_from_midnight
            + self.time_profile_points[-1].arrival_offset_from_start
            + self.time_profile_points[-1].dwell_duration,
            trip_type=trip_type,
        )

        # Sort the time profile points by arrival offset
        self.time_profile_points.sort(key=lambda x: x.arrival_offset_from_start)
        # Sort the assoc_route_stations by elapsed distance
        self.route.assoc_route_stations.sort(key=lambda x: x.elapsed_distance)

        # Create the stoptimes
        for i in range(len(self.time_profile_points)):
            if self.time_profile_points[i].station != self.route.assoc_route_stations[i].station:
                raise ValueError(
                    f"Station {self.time_profile_points[i].station.name} at position {i} does not match the station {self.route.assoc_route_stations[i].station.name} in the route"
                )

            stoptime = eflips.model.StopTime(
                scenario_id=self.route.scenario_id,
                trip=None,  # Will be done manually through append
                station=self.time_profile_points[i].station,
                arrival_time=local_midnight
                + self.start_offset_from_midnight
                + self.time_profile_points[i].arrival_offset_from_start,
                dwell_duration=self.time_profile_points[i].dwell_duration,
            )
            trip.stop_times.append(stoptime)

        return trip


def create_route(
    scenario_id: int,
    departure_station: eflips.model.Station,
    arrival_station: eflips.model.Station,
    line: eflips.model.Line,
    distance: float,
    first_grid_point: Linienfahrplan.StreckennetzDaten.Netzpunkte.Netzpunkt,
    last_grid_point: Linienfahrplan.StreckennetzDaten.Netzpunkte.Netzpunkt,
) -> eflips.model.Route:
    """
    Create a route object, setting the name up from the station

    :param scenario_id: A scenario ID
    :param departure_station: The departure station
    :param arrival_station: The arrival station
    :param line: The line object
    :param distance: The distance of the route
    :param first_grid_point: The first grid point of the route. Used to check if it is an Einsetzpunkt
    :param last_grid_point: The last grid point of the route. Used to check if it is an Aussetzpunkt
    :return: A Route object, which is not yet added to the database
    """

    name_short = line.name + " " + departure_station.name_short + " → " + arrival_station.name_short

    name = line.name + " "
    if first_grid_point.netzpunkttyp == NetzpunktNetzpunkttyp.EPKT:
        name += "Einsetzfahrt "
    elif last_grid_point.netzpunkttyp == NetzpunktNetzpunkttyp.APKT:
        name += "Aussetzfahrt "
    elif (
        first_grid_point.netzpunkttyp == NetzpunktNetzpunkttyp.BPUNKT
        or last_grid_point.netzpunkttyp == NetzpunktNetzpunkttyp.BPUNKT
    ):
        pass
    else:
        assert (
            first_grid_point.netzpunkttyp == NetzpunktNetzpunkttyp.HST
            and last_grid_point.netzpunkttyp == NetzpunktNetzpunkttyp.HST
        ), "Check what the other grid point types might mean"
    name += departure_station.name + " → " + arrival_station.name

    return eflips.model.Route(
        scenario_id=scenario_id,
        departure_station=departure_station,
        arrival_station=arrival_station,
        line=line,
        name=name,
        name_short=name_short,
        headsign=None,
        distance=distance,
        geom=None,
    )


def fix_times(points: List[TimeProfile.TimeProfilePoint]) -> List[TimeProfile.TimeProfilePoint]:
    """
    Take a list of time profile points that might have a identical arrival offsets and fix them by
    taking each set of identical arrival offsets and adding a second to each of them
    """
    # Check if the points are sorted by arrival offset
    for i in range(len(points) - 1):
        if points[i].arrival_offset_from_start > points[i + 1].arrival_offset_from_start:
            raise ValueError("Points are not sorted by arrival offset")

    # Check if there are any identical arrival offsets and fix them
    # Taking into account that there might be multiple identical arrival offsets in a row
    last_arrival_offset = None
    current_increment = 0
    for i in range(len(points)):
        current_arrival_offset = points[i].arrival_offset_from_start
        if last_arrival_offset is not None and current_arrival_offset == last_arrival_offset:
            current_increment += 1
            points[i].arrival_offset_from_start += timedelta(seconds=current_increment)

            # If we are shifting the last point back, instead we need to shift all the points forward which we
            # have shifted back before
            if i == len(points) - 1:
                for j in range(i - current_increment, i + 1):
                    points[j].arrival_offset_from_start -= timedelta(seconds=current_increment)

        else:
            current_increment = 0
        last_arrival_offset = current_arrival_offset

    # Debugging: make sure they are now strictly sorted
    # Check if the points are sorted by arrival offset
    for i in range(len(points) - 1):
        if points[i].arrival_offset_from_start >= points[i + 1].arrival_offset_from_start:
            raise ValueError("Points are not sorted by arrival offset")
    if points[-1].arrival_offset_from_start.total_seconds() % 60 != 0:
        raise ValueError("The last point does not have an arrival offset in full minutes, which it should have")

    return points


def create_routes_and_time_profiles(
    schedule: Linienfahrplan, scenario_id: int, session: Session
) -> Tuple[Dict[int, Dict[int, List[TimeProfile.TimeProfilePoint]]], Dict[int, None | eflips.model.Route]]:
    """
    First method to be used when importing a set of xml files. It takes the parsed xml data and creates the stations
    from the 'Linienfahrplan/LinienDaten/Linie/RoutenDaten/Route' entries.

    This method
    1. Adds the routes (with their corresponding AssocRouteStation objects) to the database
    2. goes through the 'Linienfahrplan/LinienDaten/Linie/RoutenDaten/Route/Fahrzeitprofil' entries and turns it into
       a list of time profiles to be used when creating the trips.
    3. Goes through the 'Linienfahrplan/Liniendaten/Fahrt' entries and assigns one of our time profiles to each trip


    :param schedule: A parsed Linienfahrplan object
    :param scenario_id: The scenario ID to use
    :param session: An open database session
    :return: A tuple of two dictionaries:
             - one contains the points of the time profiles, by their route.lfd_nr in the schedule and then by the
               fahrzeitprofil_nummer
             - the other contains the eflip route objects, by their route.lfd_nr in the schedule
    """
    logger = logging.getLogger(__name__)

    grid_points, segments, route_datas, route_lfd_nrs = setup_working_dictionaries(schedule)

    # Create the line object, if it does not exist yet
    db_line = add_or_ret_line(scenario_id, schedule.linien_daten.linie.kurzname, session)

    # We need to make sure there is only one "Linie" object, otherwise the route numbers are non-unique
    # If this turns into a list, we need to change the code below
    assert isinstance(schedule.linien_daten.linie, Linienfahrplan.LinienDaten.Linie)

    trip_time_profiles: Dict[
        int, Dict[int, List[TimeProfile.TimeProfilePoint]]
    ] = {}  # Will be keyed by route.lfd_nr, then by fahrzeitprofil_nummer
    db_routes_by_lfd_nr: Dict[int, None | eflips.model.Route] = {}  # Will be keyed by route.lfd_nr
    for route in schedule.linien_daten.linie.routen_daten.route:
        # Contrary to the naive approach, we first create the AssocRouteStation objects, and then the route object
        # This way, we are sure the departure and arrival stations match as well as the total distance
        assocs: List[eflips.model.AssocRouteStation] = []
        elapsed_distance = 0.0

        time_profile_points: Dict[
            int, List[TimeProfile.TimeProfilePoint]
        ] = {}  # Order: time profile number, time profile
        elapsed_time: Dict[int, timedelta] = {}
        for fahrzeitprofil in route.fahrzeitprofile.fahrzeitprofil:
            time_profile_points[fahrzeitprofil.fahrzeitprofil_nummer] = []
            elapsed_time[fahrzeitprofil.fahrzeitprofil_nummer] = timedelta(seconds=0)

        for i in range(len(route.punktfolge.punkt)):
            point = route.punktfolge.punkt[i]
            # Load data to be used later
            station = add_or_ret_station_for_grid_point(scenario_id, point.netzpunkt, grid_points, session)
            grid_point = grid_points[point.netzpunkt]
            geom = soldner_to_pointz(grid_point.xkoordinate, grid_point.ykoordinate)

            # Temporal: Update driving times
            driving_times: Dict[int, Tuple[timedelta, timedelta]] = {}  # Order: driving, waiting
            for fahrzeitprofil in route.fahrzeitprofile.fahrzeitprofil:
                driving_time_point = fahrzeitprofil.fahrzeitprofilpunkte.punkt[i]
                driving_times[fahrzeitprofil.fahrzeitprofil_nummer] = (
                    timedelta(seconds=driving_time_point.streckenfahrzeit),
                    timedelta(seconds=driving_time_point.wartezeit),
                )
                elapsed_time[fahrzeitprofil.fahrzeitprofil_nummer] += timedelta(
                    seconds=driving_time_point.streckenfahrzeit
                )

            # Geographic: Update elapsed distance
            if i > 0:
                segment_id = route.streckenfolge.strecke[i - 1].strecken_id
                segment = segments[segment_id]
                elapsed_distance += segment.streckenlaenge

            # SPECIAL FIXES
            # Some routes have a distance of zero even once the last point is reached
            if i == len(route.punktfolge.punkt) - 1 and elapsed_distance == 0:
                # We mark these by putting an obscenely large number in the distance
                logger.warning(f"Route {route.lfd_nr} of line {db_line.name} has a zero distance at the end.")
                elapsed_distance += 1e6 * 1000  # One million kilometers

            # Some time profiles have a zero time at the end
            if i == len(route.punktfolge.punkt) - 1:
                for fahrzeitprofil in route.fahrzeitprofile.fahrzeitprofil:
                    if elapsed_time[fahrzeitprofil.fahrzeitprofil_nummer] == timedelta(seconds=0):
                        logger.info(
                            f"Route {route.lfd_nr} of line {db_line.name} has a zero time at the end. Calculating duration with a fixed speed of 30 km/h."
                        )

                        SPEED = 30 / 3.6  # 30 km/h in m/s
                        if elapsed_distance != 0:
                            duration = elapsed_distance / SPEED
                        else:
                            logger.warning(
                                f"Route {route.lfd_nr} of line {db_line.name} has a zero distance at the end. Calculating duration with a fixed speed of 30 km/h and a distance of 1 km."
                            )
                            duration = 1000 / SPEED

                        # Now, depending on whether it is an EInsetzfahrt or Aussetzfahrt, we shoft the beginning forward
                        # or the end backward
                        first_grid_point = grid_points[route.punktfolge.punkt[0].netzpunkt]
                        last_grid_point = grid_points[route.punktfolge.punkt[-1].netzpunkt]
                        if first_grid_point.netzpunkttyp == NetzpunktNetzpunkttyp.EPKT:
                            # Shift the first entry back by the duration
                            time_profile_points[fahrzeitprofil.fahrzeitprofil_nummer][
                                0
                            ].arrival_offset_from_start = timedelta(seconds=-duration)
                        elif last_grid_point.netzpunkttyp == NetzpunktNetzpunkttyp.APKT:
                            elapsed_time[fahrzeitprofil.fahrzeitprofil_nummer] += timedelta(seconds=duration)
                        else:
                            raise ValueError(
                                f"Route {route.lfd_nr} of line {db_line.name} has a zero time at the end, but is neither an Einsetzfahrt nor an Aussetzfahrt"
                            )

            # We always add the first point
            if i == 0:
                # Geographic
                assoc = eflips.model.AssocRouteStation(
                    scenario_id=scenario_id,
                    route=None,  # Will be set later
                    station=station,
                    location=geom,
                    elapsed_distance=elapsed_distance,
                )
                assocs.append(assoc)

                # Temporal
                # Here, the time driven until this point must be 0
                for fahrzeitprofil in route.fahrzeitprofile.fahrzeitprofil:
                    assert driving_times[fahrzeitprofil.fahrzeitprofil_nummer][0] == timedelta(seconds=0)
                    waiting_time = driving_times[fahrzeitprofil.fahrzeitprofil_nummer][1]
                    time_profile_points[fahrzeitprofil.fahrzeitprofil_nummer].append(
                        TimeProfile.TimeProfilePoint(
                            station=station,
                            arrival_offset_from_start=timedelta(seconds=0),
                            dwell_duration=waiting_time,
                        )
                    )
            # Otherwise, we only add a point if the station and elapsed distance are different from the previous point
            elif station != assocs[-1].station and elapsed_distance != assocs[-1].elapsed_distance:
                # Geographic
                assoc = eflips.model.AssocRouteStation(
                    scenario_id=scenario_id,
                    route=None,  # Will be set later
                    station=station,
                    location=geom,
                    elapsed_distance=elapsed_distance,
                )
                assocs.append(assoc)

                # Temporal
                for fahrzeitprofil in route.fahrzeitprofile.fahrzeitprofil:
                    time_profile_points[fahrzeitprofil.fahrzeitprofil_nummer].append(
                        TimeProfile.TimeProfilePoint(
                            station=station,
                            arrival_offset_from_start=elapsed_time[fahrzeitprofil.fahrzeitprofil_nummer],
                            dwell_duration=timedelta(seconds=0),
                        )
                    )
            # If we are at the last point, we will need to update the elapsed distance and time
            # even if we would not normally add the point
            elif i == len(route.punktfolge.punkt) - 1:
                # Geographic
                assocs[-1].elapsed_distance = elapsed_distance

                # Temporal
                for fahrzeitprofil in route.fahrzeitprofile.fahrzeitprofil:
                    time_profile_points[fahrzeitprofil.fahrzeitprofil_nummer][
                        -1
                    ].arrival_offset_from_start = elapsed_time[fahrzeitprofil.fahrzeitprofil_nummer]

            # In the end, add the waiting time at this point (if any)
            # to the elapsed_time and the dwell_durations of the last time profile point
            for fahrzeitprofil in route.fahrzeitprofile.fahrzeitprofil:
                waiting_time = driving_times[fahrzeitprofil.fahrzeitprofil_nummer][1]
                elapsed_time[fahrzeitprofil.fahrzeitprofil_nummer] += waiting_time
                time_profile_points[fahrzeitprofil.fahrzeitprofil_nummer][-1].dwell_duration += waiting_time

        del (
            point,
            station,
            grid_point,
            geom,
            driving_time_point,
            segment_id,
            segment,
            waiting_time,
            assoc,
        )  # Clean up the debugger

        # Now, do some sanity checks
        last_distance = None
        if len(assocs) < 2:
            # There are some routes which we have manually checked out and figured to be pointless
            if [p.netzpunkt for p in route.punktfolge.punkt] == [102001974, 101001974, 101029999, 101001974, 102001974]:
                # This is a bus which just stands aroung in Alt-Gatow for a while ?!?!?!?!?
                logger.info(f"Route {route.lfd_nr} of line {db_line.name} is a pointless route. Skipping")
                # Write none to the dict of routes, to show we're aware of this route, but it's pointless
                db_routes_by_lfd_nr[route.lfd_nr] = None
                continue
            elif [p.netzpunkt for p in route.punktfolge.punkt] == [102021010, 101021010, 101002083, 102002083]:
                # This might be a turning-around at Osloer Straße. We don't need it
                logger.info(f"Route {route.lfd_nr} of line {db_line.name} is a pointless route. Skipping")
                # Write none to the dict of routes, to show we're aware of this route, but it's pointless
                db_routes_by_lfd_nr[route.lfd_nr] = None
                continue
            elif [p.netzpunkt for p in route.punktfolge.punkt] == [102004107, 101004107, 101004108, 102004108]:
                # Turning around at Hermannstraße
                logger.info(f"Route {route.lfd_nr} of line {db_line.name} is a pointless route. Skipping")
                # Write none to the dict of routes, to show we're aware of this route, but it's pointless
                db_routes_by_lfd_nr[route.lfd_nr] = None
                continue
            raise ValueError("There should be at least one assoc")
        for assoc in assocs:
            if last_distance is not None:
                if not assoc.elapsed_distance > last_distance:
                    raise ValueError("The elapsed distance should be increasing for each assoc")
            last_distance = assoc.elapsed_distance
        del last_distance

        for fahrzeitprofil in route.fahrzeitprofile.fahrzeitprofil:
            last_time = None
            this_vehicles_time_profile_points = time_profile_points[fahrzeitprofil.fahrzeitprofil_nummer]
            if len(this_vehicles_time_profile_points) < 2:
                raise ValueError("There should be at least one time profile point")
            for time_profile_point in this_vehicles_time_profile_points:
                if last_time is not None:
                    if not time_profile_point.arrival_offset_from_start > last_time:
                        logger.info(
                            f"Route {route.lfd_nr} of line {db_line.name} has a time profile with a non-increasing arrival offset. Fixing them by adding one second to each identical arrival offset"
                        )
                        this_vehicles_time_profile_points = fix_times(this_vehicles_time_profile_points)
                last_time = time_profile_point.arrival_offset_from_start + time_profile_point.dwell_duration
        del this_vehicles_time_profile_points, time_profile_point, last_time

        # Save the time profile points - we do this even if the route is pointless, as there might be trips on it
        trip_time_profiles[route.lfd_nr] = time_profile_points

        # We can now create the route object and see if it already exists
        db_route = create_route(
            scenario_id=scenario_id,
            departure_station=assocs[0].station,
            arrival_station=assocs[-1].station,
            line=db_line,
            distance=elapsed_distance,
            first_grid_point=grid_points[route.punktfolge.punkt[0].netzpunkt],
            last_grid_point=grid_points[route.punktfolge.punkt[-1].netzpunkt],
        )

        # Now we can check if there already is a route with the same assocs
        route_already_exists = False
        route_q = (
            session.query(eflips.model.Route)
            .filter(eflips.model.Route.scenario_id == scenario_id)
            .filter(eflips.model.Route.name == db_route.name)
            .filter(eflips.model.Route.name_short == db_route.name_short)
        )
        if route_q.count() > 0:
            for existing_route in route_q:
                if len(existing_route.assoc_route_stations) == len(assocs):
                    equal = True
                    for i in range(len(existing_route.assoc_route_stations)):
                        if (
                            existing_route.assoc_route_stations[i].station != assocs[i].station
                            or existing_route.assoc_route_stations[i].location != assocs[i].location
                            or existing_route.assoc_route_stations[i].elapsed_distance != assocs[i].elapsed_distance
                        ):
                            equal = False
                            break  # We can stop comparing the assocs
                    if equal:
                        # If the route already exist, we will return it as the one to create the trips for
                        db_routes_by_lfd_nr[route.lfd_nr] = existing_route
                        route_already_exists = True
                        break  # We can stop comparing the other routes
        if not route_already_exists:
            session.add(db_route)
            for assoc in assocs:
                assoc.route = db_route
                session.add(assoc)
            # Also add it to the dict of routes
            db_routes_by_lfd_nr[route.lfd_nr] = db_route

        del (
            assocs,
            elapsed_distance,
            elapsed_time,
            time_profile_points,
            route_already_exists,
        )

    return trip_time_profiles, db_routes_by_lfd_nr


def create_trip_prototypes(
    schedule: Linienfahrplan,
    route_time_profileps: Dict[int, Dict[int, List[TimeProfile.TimeProfilePoint]]],
    db_routes_by_lfd_nr: Dict[int, None | eflips.model.Route],
) -> Dict[int, None | TimeProfile]:
    """
    From a list of time profiles, create a list of trip prototypes, one for each trip in the schedule.

    :param schedule: A parsed Linienfahrplan object
    :param route_time_profileps: A Dict of all the time profiles for each route, ordered by the route.lfd_nr (not to be
           confused with routenvatiante.lfd_nr)
    :return: A Dict with time profiles, ordered by the trip.ID (Fahrt.ID)
    """
    logger = logging.getLogger(__name__)

    grid_points, segments, route_datas, route_lfd_nrs = setup_working_dictionaries(schedule)

    # We need to make sure there is only one "Linie" object, otherwise the route numbers are non-unique
    # If this turns into a list, we need to change the code below
    assert isinstance(schedule.linien_daten.linie, Linienfahrplan.LinienDaten.Linie)

    time_profiles_by_trip_id: Dict[int, None | TimeProfile] = {}
    for fahrt in schedule.fahrt_daten.fahrt:
        route_lfd_nr = route_lfd_nrs[
            fahrt.lfd_nr_routenvariante
        ]  # Convert the routenvariante.lfd_nr to the route.lfd_nr
        db_route = db_routes_by_lfd_nr[route_lfd_nr]
        if db_route is None:
            # That means that thios route was seen by the route creation function, but was deemed pointless
            logger.info(f"Refusing to create trip prototype for trip {fahrt.id}, as the route is pointless")

            # Put an explicit None in the dict, so we know we've seen this trip
            time_profiles_by_trip_id[fahrt.id] = None

            continue
        time_profile = route_time_profileps[route_lfd_nr][fahrt.fahrzeitprofil]

        for i in range(len(time_profile)):
            if time_profile[i].station != db_route.assoc_route_stations[i].station:
                raise ValueError(
                    f"Station {time_profile[i].station.name} at position {i} does not match the station {db_route.assoc_route_stations[i].station.name} in the route"
                )

        profile = TimeProfile(
            route=db_route,
            start_offset_from_midnight=timedelta(seconds=fahrt.startzeit),
            time_profile_points=time_profile,
        )
        time_profiles_by_trip_id[fahrt.id] = profile
    return time_profiles_by_trip_id


def create_trips_and_vehicle_schedules(
    schedule: Linienfahrplan, trip_prototypes: Dict[int, None | TimeProfile], scenario_id: int, session: Session
) -> None:
    """
    Creates the trips and vehicle schedules from the parsed Linienfahrplan object
    :param schedule: A parsed Linienfahrplan object
    :param trip_prototypes: A dictionary containing *all* the available time profiles, by their trip.ID. This (rather
           than the *schedule's* time profiles) is used, because the schedule might contain trips that are not in the
           schedule (if a vehicle from this line goes to another line)
    :param scenario_id: The scenario ID to use
    :param session: An open database session
    :return: Nothing. The trips are added to the database
    """
    logger = logging.getLogger(__name__)

    for fahrzeugumlauf in schedule.fahrzeugumlauf_daten.fahrzeugumlauf:
        # The Fharzeugumlauf has an Umlaeufe object, which implies there could be multiple.
        # We do not support this
        vehicle_type = add_or_ret_vehicle_type(scenario_id, fahrzeugumlauf.fahrzeugtyp, session)
        rotation = eflips.model.Rotation(
            scenario_id=scenario_id,
            id=None,
            name="",
            vehicle_type=vehicle_type,
            allow_opportunity_charging=True,
        )
        session.add(rotation)
        rotation_trips = []
        for xml_rotation in fahrzeugumlauf.umlaeufe.umlauf:
            date_german = xml_rotation.kalenderdatum.split(".")
            the_date = date(day=int(date_german[0]), month=int(date_german[1]), year=int(date_german[2]))
            rotation.name += f"{xml_rotation.umlaufbezeichnung} "
            # Assemble the list of trips
            for umlaufteilgruppe in xml_rotation.umlaufteilgruppen.umlaufteilgruppe:
                if umlaufteilgruppe.fahrtreihenfolge is None:
                    logger.info(
                        f"Umlauf {xml_rotation.umlauf_id} has a Umlaufteilgruppe without a Fahrtreihenfolge. Skipping"
                    )
                    continue
                # Make sure the vehicle type stays the same
                part_trips = []
                for vehicle_type_str in umlaufteilgruppe.fahrzeugtyp:
                    if vehicle_type_str != vehicle_type.name_short:
                        raise ValueError(f"Vehicle type changed while within Umlauf {xml_rotation.umlauf_id}!")
                    for fahrt in umlaufteilgruppe.fahrtreihenfolge.fahrt:
                        if fahrt.fahrt_id not in trip_prototypes.keys():
                            raise ValueError(
                                f"Trip {fahrt.fahrt_id} from Umlauf {xml_rotation.umlauf_id} is not known! "
                                f"Is it from another Linie, which XML file we are missing?"
                            )
                        tp = trip_prototypes[fahrt.fahrt_id]
                        if tp is None:
                            logger.info(
                                f"Trip {fahrt.fahrt_id} from Umlauf {xml_rotation.umlauf_id} is from a pointless route. Skipping"
                            )
                            continue
                        with session.no_autoflush:  # Get rid of spurious SAWarnings
                            part_trips.append(tp.to_trip(rotation, the_date))
                    rotation_trips.extend(part_trips)

            ### SPECIAL FIXES
            # Do some cleanup. There may be geographical inconsistencies caused by a trip's stop being different from the
            # stop of the previous trip. This might be caused by a station having multiple "Haltestellenbereiche"
            # We see if the first four letters of the stations short name are the same, and if so, we change the latter
            # trip's departure station to the former trip's arrival station
            for i in range(len(rotation_trips) - 1):
                cur_trip = rotation_trips[i]
                next_trip = rotation_trips[i + 1]
                if cur_trip.stop_times[-1].station != next_trip.stop_times[0].station:
                    logger.info(
                        f"Trip {cur_trip.id} and {next_trip.id} have different stations: first {cur_trip.stop_times[-1].station.name} (ID f{cur_trip.stop_times[-1].station.id}), then {next_trip.stop_times[0].station.name} (ID f{next_trip.stop_times[0].station.id})."
                    )
            session.add_all(rotation_trips)

        rotation.name = rotation.name.strip()


def recenter_station(station: eflips.model.Station, session: Session) -> None:
    """
    Puts a station's location at the median of it's associations
    :param station:
    :param session:
    :return: Nothing. The station is updated in the database
    """
    # For each association, load the location and add it to a list
    xs, ys, zs = [], [], []
    for assoc in station.assoc_route_stations:
        if isinstance(assoc.location, str):
            loc_str = assoc.location.lstrip("SRID=4326;POINTZ(").rstrip(")")
            x, y, z = loc_str.split(" ")
            x_f, y_f, z_f = float(x), float(y), float(z)
            xs.append(x_f)
            ys.append(y_f)
            zs.append(z_f)
        else:
            assert isinstance(assoc.location, WKBElement)
            shape = to_shape(assoc.location)  # typ
            xs.append(shape.x)
            ys.append(shape.y)
            zs.append(shape.z)

    # Calculate the median of the list
    median_x = statistics.median(xs)
    median_y = statistics.median(ys)
    median_z = statistics.median(zs)

    # Create a new location from the median
    new_location = f"SRID=4326;POINTZ({median_x} {median_y} {median_z})"

    # Update the station
    station.geom = new_location  # type: ignore


def fix_max_sequence(database_url: str) -> None:
    """
    Run some SQL in order to set the nextval() sequence to the maximum value of the id column for each table.
    :param database_url: The database URL to use
    :return: None
    """
    SEQUENCES = [
        "Scenario_id_seq",
        "Plan_id_seq",
        "Process_id_seq",
        "BatteryType_id_seq",
        "VehicleClass_id_seq",
        "Line_id_seq",
        "Station_id_seq",
        "Depot_id_seq",
        "AssocPlanProcess_id_seq",
        "VehicleType_id_seq",
        "Route_id_seq",
        "Area_id_seq",
        "Vehicle_id_seq",
        "AssocVehicleTypeVehicleClass_id_seq",
        "AssocRouteStation_id_seq",
        "AssocAreaProcess_id_seq",
        "Rotation_id_seq",
        "Trip_id_seq",
        "Event_id_seq",
        "StopTime_id_seq",
    ]
    conn = psycopg2.connect(database_url)
    conn.autocommit = True
    for sequence in SEQUENCES:
        table_name = sequence.split("_")[0]
        key_name = sequence.split("_")[1]
        with conn.cursor() as cur:
            cur.execute(f'SELECT MAX("{key_name}") FROM "{table_name}"')
            res = cur.fetchone()
            max_id = res[0] if res is not None else None
            if max_id is not None:
                cur.execute(f'ALTER SEQUENCE "{sequence}" RESTART WITH {max_id + 1}')
                cur.execute(f'SELECT NEXTVAL(\'"public"."{sequence}"\')')
                res = cur.fetchone()
                new_max_id = res[0] if res is not None else None
                if new_max_id <= max_id:
                    raise ValueError(f"Sequence {sequence} did not restart properly. It is still at {new_max_id}")
    conn.close()


def merge_identical_stations(scenario_id: int, session: Session) -> None:
    """
    This method merges stations which have an identical name.
    :param scenario_id: The scenario ID to use
    :param session: An open database session
    :return: Nothing. The stations are updated in the database
    """
    station_names = (
        session.query(eflips.model.Station.name, func.count())
        .filter(eflips.model.Station.scenario_id == scenario_id)
        .group_by(eflips.model.Station.name)
        .all()
    )
    for station_name, count in station_names:
        if count == 1:
            continue
        all_stations_for_name = (
            session.query(eflips.model.Station)
            .filter(eflips.model.Station.name == station_name)
            .filter(eflips.model.Station.scenario_id == scenario_id)
            .order_by(eflips.model.Station.id)
            .all()
        )
        # Keep the ID of the first station
        first_station_id = all_stations_for_name[0].id
        # For the Other IDs, update all objects that reference them to reference the first station
        for station in all_stations_for_name[1:]:
            # Update the routes
            session.query(eflips.model.Route).filter(eflips.model.Route.departure_station_id == station.id).update(
                {"departure_station_id": first_station_id}
            )
            session.query(eflips.model.Route).filter(eflips.model.Route.arrival_station_id == station.id).update(
                {"arrival_station_id": first_station_id}
            )

            # Update the AssocRouteStations
            session.query(eflips.model.AssocRouteStation).filter(
                eflips.model.AssocRouteStation.station_id == station.id
            ).update({"station_id": first_station_id})

            # Update the stop times
            session.query(eflips.model.StopTime).filter(eflips.model.StopTime.station_id == station.id).update(
                {"station_id": first_station_id}
            )

            # Delete the station
            session.refresh(station)  # Refresh the station object to get the updated routes
            session.delete(station)


def merge_identical_rotations(scenario_id: int, session: Session) -> None:
    """
    This method merges rotations which have an identical name. It does so by merging everything from
    a rotation beginning at a "Betriebshof"to the next "Betriebshof".
    This needs to be done because different parts of a rotation may be in different files, and so different
    "Rotation" objects for what is essentially the same rotation may be created.
    :param scenario_id: the scenario ID to use
    :param session: an open database session
    :return: Nothing. The rotations are updated in the database
    """
    DEPOT_NAME = "Betriebshof"

    rotation_names = (
        session.query(eflips.model.Rotation.name)
        .filter(eflips.model.Rotation.scenario_id == scenario_id)
        .group_by(eflips.model.Rotation.name)
        .all()
    )
    for rotation_name in rotation_names:
        all_rots_for_name = (
            session.query(eflips.model.Rotation)
            .filter(eflips.model.Rotation.name == rotation_name[0])
            .filter(eflips.model.Rotation.scenario_id == scenario_id)
            .all()
        )
        # Order them by the first trip's departure time
        all_rots_for_name = sorted(all_rots_for_name, key=lambda rot: rot.trips[0].departure_time)

        list_of_rotation_id_tuples_to_merge: List[List[int]] = []
        rotation_ids_to_merge: List[int] = []

        for rotation in all_rots_for_name:
            first_station_name = rotation.trips[0].route.departure_station.name
            last_station_name = rotation.trips[-1].route.arrival_station.name
            first_station_departure_time = rotation.trips[0].departure_time
            last_station_arrival_time = rotation.trips[-1].arrival_time

            if DEPOT_NAME in first_station_name and DEPOT_NAME in last_station_name:
                # This is a rotation that starts and ends at the depot
                # We don't need to merge it with anything
                continue
            elif DEPOT_NAME in first_station_name and DEPOT_NAME not in last_station_name:
                # This rotation starts at the depot and ends somewhere else
                # Start a new list after appending the current list to the list of lists
                list_of_rotation_id_tuples_to_merge.append(rotation_ids_to_merge)
                rotation_ids_to_merge = [rotation.id]
            elif DEPOT_NAME not in first_station_name and DEPOT_NAME in last_station_name:
                # This rotation starts somewhere else and ends at the depot
                # Append the current rotation to the list
                rotation_ids_to_merge.append(rotation.id)
                list_of_rotation_id_tuples_to_merge.append(rotation_ids_to_merge)
                rotation_ids_to_merge = []
            elif DEPOT_NAME not in first_station_name and DEPOT_NAME not in last_station_name:
                # This rotation starts and ends somewhere else
                # Append the current rotation to the list
                rotation_ids_to_merge.append(rotation.id)
            else:
                raise ValueError("This should never happen")

        # Remove empty lists
        list_of_rotation_id_tuples_to_merge = [x for x in list_of_rotation_id_tuples_to_merge if len(x) > 0]

        # Go through the list of lists and merge the rotations, if they form a valid rotation
        for rotation_ids_to_merge in list_of_rotation_id_tuples_to_merge:
            rotations = (
                session.query(eflips.model.Rotation).filter(eflips.model.Rotation.id.in_(rotation_ids_to_merge)).all()
            )
            # Order the rotations by the first trip's departure time
            rotations = sorted(rotations, key=lambda rot: rot.trips[0].departure_time)

            # Check whether the rotations can be merged
            # The first station of the first rotation must be the same as the last station of the last rotation
            # And both must contain "Betriebshof"
            if (
                rotations[0].trips[0].route.departure_station.name != rotations[-1].trips[-1].route.arrival_station.name
                or "Betriebshof" not in rotations[0].trips[0].route.departure_station.name
                or "Betriebshof" not in rotations[-1].trips[-1].route.arrival_station.name
            ):
                # If the merge is not possible we delete these rotations
                for rotation in rotations:
                    for trip in rotation.trips:
                        for stop_time in trip.stop_times:
                            session.delete(stop_time)
                        session.delete(trip)
                    session.delete(rotation)
            else:
                # Merge the rotations by creating a new rotation
                new_rotation = eflips.model.Rotation(
                    name=rotations[0].name,
                    scenario_id=rotations[0].scenario_id,
                    vehicle_type_id=rotations[0].vehicle_type_id,
                    allow_opportunity_charging=rotations[0].allow_opportunity_charging,
                )
                session.add(new_rotation)
                session.flush()
                for rotation in rotations:
                    session.query(eflips.model.Trip).filter(eflips.model.Trip.rotation_id == rotation.id).update(
                        {"rotation_id": new_rotation.id}
                    )
                    session.refresh(
                        rotation
                    )  # Refresh the rotation object to get the updated trips (there are now none)
                    session.delete(rotation)
                    session.flush()


def identify_and_delete_overlapping_rotations(scenario_id: int, session: Session) -> None:
    """
    Identify and delete rotations with overlapping trips. These may be caused by the rotation merging function.
    :param scenario_id: The scenario ID to use
    :param session: An open database session
    :return: Nothing. The rotations are updated in the database
    """
    logger = logging.getLogger(__name__)
    rotations = session.query(eflips.model.Rotation).filter(eflips.model.Rotation.scenario_id == scenario_id).all()
    for rotation in rotations:
        trips = rotation.trips
        for i in range(len(trips) - 1):
            if trips[i].arrival_time > trips[i + 1].departure_time:
                logger.warning(
                    f"Rotation {rotation.id} has overlapping trips {trips[i].id} and {trips[i + 1].id}. Deleting the rotation"
                )
                for trip in trips:
                    for stop_time in trip.stop_times:
                        session.delete(stop_time)
                    session.delete(trip)
                session.delete(rotation)
                break


def ingest_bvgxml(
    paths: Union[str, List[str]],
    database_url: str,
    clear_database: bool = False,
    multithreading: bool = True,
    log_level: str = "WARNING",
) -> None:
    """
    The main method for ingesting BVG-XML format files into the database.

    :param paths: Either a directory or a list of files to ingest
    :param database_url: The database URL to use for the ingestion
    :param clear_database: Whether to clear the database and create a new schema before ingesting
    :param multithreading: Whether to use multithreading or not. Useful to disable for debugging
    :return:
    """

    match log_level:
        case "DEBUG":
            logging.basicConfig(level=logging.DEBUG)
        case "INFO":
            logging.basicConfig(level=logging.INFO)
        case "WARNING":
            logging.basicConfig(level=logging.WARNING)
        case "ERROR":
            logging.basicConfig(level=logging.ERROR)
        case "CRITICAL":
            logging.basicConfig(level=logging.CRITICAL)
        case _:
            raise ValueError("Invalid log level. Must be one of DEBUG, INFO, WARNING, ERROR, CRITICAL")

    logger = logging.getLogger(__name__)

    if isinstance(paths, str):
        if os.path.isdir(paths):
            # Find all xml files in the directory
            paths = glob.glob(os.path.join(paths, "*.xml"))
        else:
            paths = [paths]
    paths_pathlike = [Path(p) for p in paths]

    TOTAL_STEPS = 11

    ### STEP 1: Load the XML files into memory
    # First, we go through all the files and load them into memory
    schedules = []
    if multithreading:
        with Pool() as pool:
            for schedule in tqdm(
                pool.imap_unordered(load_and_validate_xml, paths_pathlike),
                total=len(paths_pathlike),
                desc=f"(1/{TOTAL_STEPS}) Loading XML files",
            ):
                schedules.append(schedule)
    else:
        schedules = []
        for path in tqdm(paths_pathlike, desc=f"(1/{TOTAL_STEPS}) Loading XML files"):
            schedules.append(load_and_validate_xml(path))

    ### STEP 1.5: Create the database session and scenario
    engine = create_engine(database_url)
    if clear_database:
        eflips.model.Base.metadata.drop_all(engine)
        eflips.model.setup_database(engine)
    session = Session(engine)
    scenario = eflips.model.Scenario(
        name=f"Created by BVG-XML Ingestion on {socket.gethostname()} at {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}"
    )
    session.add(scenario)
    session.flush()
    scenario_id = scenario.id

    ### STEP 2: Create the stations
    # Now, we go through the schedules and create the stations
    # No multithreading, because that would just create duplicate stations
    for schedule in tqdm(schedules, desc=f"(2/{TOTAL_STEPS}) Creating stations"):
        create_stations(schedule, scenario_id, session)

    ### STEP 3: Create the routes and save some data for later
    # Again no multithreading
    create_route_results: List[
        Tuple[
            Linienfahrplan,
            Dict[int, Dict[int, List[TimeProfile.TimeProfilePoint]]],
            Dict[int, None | eflips.model.Route],
        ]
    ] = []
    for schedule in tqdm(schedules, desc=f"(3/{TOTAL_STEPS}) Creating routes"):
        trip_time_profiles, db_routes_by_lfd_nr = create_routes_and_time_profiles(schedule, scenario_id, session)
        create_route_results.append((schedule, trip_time_profiles, db_routes_by_lfd_nr))

    ### STEP 4: Create the trip prototypes
    # This can be done in parallel, but we don't need to do it, it's fast enough
    all_trip_protoypes: List[Dict[int, None | TimeProfile]] = []
    for create_route_result in tqdm(create_route_results, desc=f"(4/{TOTAL_STEPS}) Creating trip prototypes"):
        trip_prototypes = create_trip_prototypes(create_route_result[0], create_route_result[1], create_route_result[2])
        all_trip_protoypes.append(trip_prototypes)

    # Unify the dictionaries, making sure the contents are the same if there is a duplicate key
    trip_prototypes = {}
    for the_dict in all_trip_protoypes:
        for fahrt_id, time_profile in the_dict.items():
            if fahrt_id in trip_prototypes:
                if trip_prototypes[fahrt_id] != time_profile:
                    raise ValueError(f"Trip {fahrt_id} has two different time profiles in different schedules")
            else:
                trip_prototypes[fahrt_id] = time_profile

    ### STEP 5: Create the trips and vehicle schedules
    for schedule in tqdm(schedules, desc=f"(5/{TOTAL_STEPS}) Creating trips and vehicle schedules"):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=ConsistencyWarning)
            create_trips_and_vehicle_schedules(schedule, trip_prototypes, scenario_id, session)

    ### STEP 6: Set the geom of the stations
    # No multithreading, because it should be fast enough
    stations_without_geom_q = (
        session.query(eflips.model.Station)
        .join(eflips.model.AssocRouteStation)
        .filter(eflips.model.Station.scenario_id == scenario_id)
        .distinct(eflips.model.Station.id)
    )
    for station in tqdm(
        stations_without_geom_q,
        desc=f"(6/{TOTAL_STEPS}) Setting station geom",
        total=stations_without_geom_q.count(),
    ):
        # Get the median of the assoc_route_stations
        recenter_station(station, session)

    ### STEP 7: Fix the routes with very large distances:
    # There are some routes which have a distance of zero even once the last point is reached
    # We set their distance to a very large number. Now we set it to the geometric distance between the first and last
    # point
    long_route_q = (
        session.query(eflips.model.Route)
        .filter(eflips.model.Route.scenario_id == scenario_id)
        .filter(eflips.model.Route.distance >= 1e6 * 1000)
    )
    for route in tqdm(long_route_q, desc=f"(7/{TOTAL_STEPS}) Fixing long routes", total=long_route_q.count()):
        first_point = route.departure_station.geom
        last_point = route.arrival_station.geom

        first_point_soldner = func.ST_Transform(first_point, 3068)
        last_point_soldner = func.ST_Transform(last_point, 3068)
        dist_q = ST_Distance(first_point_soldner, last_point_soldner)

        dist = session.query(dist_q).one()[0]

        with session.no_autoflush:
            route.distance = dist
            route.assoc_route_stations[-1].elapsed_distance = dist
        route.name = "CHECK DISTANCE: " + route.name

    session.commit()
    session.expunge_all()

    # STEP 8: Merge identical stations
    print(f"(8/{TOTAL_STEPS}) Merging identical stations")
    merge_identical_stations(scenario_id, session)

    session.commit()
    session.expunge_all()

    # STEP 9: Combine rotations with the same name
    print(f"(9/{TOTAL_STEPS}) Merging identical rotations")
    merge_identical_rotations(scenario_id, session)

    # STEP 10: Identify overlapping rotations
    print(f"(10/{TOTAL_STEPS}) Identifying and deleting overlapping rotations")
    identify_and_delete_overlapping_rotations(scenario_id, session)

    # Commit and close this session
    session.commit()
    session.close()

    # STEP 11: Fix the max sequence numbers
    print(f"(11/{TOTAL_STEPS}) Fixing max sequence numbers")
    fix_max_sequence(database_url)

    print(
        """
    The import is complete. You may still want to:
    - Figure out what to do with  the Rotations starting or ending at an "Abstellfläche"
    """
    )


if __name__ == "__main__":
    fire.Fire(ingest_bvgxml)



import os
import tempfile
from abc import abstractmethod, ABC
from enum import Enum
from pathlib import Path
from typing import Tuple, Dict, Callable
from uuid import UUID


class AbstractIngester(ABC):
    def __init__(self, database_url: str):
        """
        Constructor for the BaseIngester class.

        :param database_url: A string representing the URL of the database to ingest into. Must be of the form
                             postgresql://user:password@host:port/database. An url of the format
                             postgis://user:password@host:port/database (django style) will be converted to the4
        :return: None
        """
        if database_url.startswith("postgis://"):
            database_url = database_url.replace("postgis://", "postgresql://", 1)
        self.database_url = database_url

    @abstractmethod
    def prepare(
        self,
        progress_callback: None | Callable[[float], None] = None,
        **kwargs: Dict[str, str | int | float | bool | Enum | Path],
    ) -> Tuple[bool, UUID | Dict[str, str]]:
        """
        Prepare and validate the input data for ingestion.

        The keyword arguments should be set to specific arguments when subclassing this method. Additionally, the
        :meth:`prepare_param_names` and :meth:`prepare_param_description` methods should be implemented to provide hints
        for the parameters of this method.

        This method of subclassing vilotaes the Liskov Substitution Principle, as the subclassed method has a different
        signature than the superclass method. As such, the subclass should be marked with # type: ignore.

        The types for keyword arguments should be limited to the following:
        - str: For text data.
        - int: For integer data.
        - float: For floating point data.
        - bool: For boolean data.
        - subclass of Enum: For enumerated data.
        - pathlib.Path: For file paths. *This is what should be done to express the need for uploaded files.*

        Additionally the method should accept a progress_callback parameter, which is a function that accepts a float
        value between 0 and 1. This function should be called periodically to update the progress of the ingestion
        process.

        When developing a (web) interface, it is suggested to use introspection on the parameters of this method to
        generate a form for the user to fill in. This can be done by using the :meth:`inspect.signature` method from the
        `inspect` module.

        This method should validate the input data and save its working data to a temporary location (suggested to be
        the path returned by the :meth:`path_for_uuid` method). If the data is valid, it should return a UUID
        representing the data and a boolean indicating that the data is valid. If the data is invalid, it should return
        a dictionary containing the error names and messages.

        :param kwargs: A dictionary containing the input data.
        :return: A tuple containing a boolean indicating whether the input data is valid and either a UUID or a dictionary
                 containing the error message.
        """
        pass

    @abstractmethod
    def ingest(self, uuid: UUID, progress_callback: None | Callable[[float], None] = None) -> None:
        """
        Ingest the data into the database. In order for this method to be called, the :meth:`prepare` method must have
        returned a UUID, indicating that the preparation was successful.

        This method must call the progress_callback function periodically to update the progress of the ingestion
        process. The progress_callback function should accept a float value between 0 and 1.

        The method should check if a scenario with the same UUID already exists in the database. If it does, it should
        add the data to the existing scenario. If it does not, it should create a new scenario.

        :param uuid: A UUID representing the data to ingest.
        :return: Nothing. If unexpected errors occur, they should be raised as exceptions.
        """
        pass

    @classmethod
    @abstractmethod
    def prepare_param_names(cls) -> Dict[str, str | Dict[Enum, str]]:
        """
        A dictionary containing the parameter names for :meth:`prepare`.

        These should be short, descriptive names for the parameters of the :meth:`prepare` method. The keys must be the
        names of the parameters, and the values should be strings describing the parameter. If the keyword argument is
        an enumerated type, the value should be a dictionary with the keys being the members.

        This method can then be used to generate a help text for the user.

        :return: A dictionary containing the parameter hints for the prepare method.
        """
        pass

    @classmethod
    @abstractmethod
    def prepare_param_description(cls) -> Dict[str, str | Dict[Enum, str]]:
        """
        A dictionary containing the parameter descriptions for :meth:`prepare`.

        These should be longer, more detailed descriptions of the parameters of the :meth:`prepare`
        method. The keys must be the names of the parameters, and the values should be strings describing the parameter.

        This method can then be used to generate a help text for the user.

        :return: A dictionary containing the parameter hints for the prepare method.
        """
        pass

    def path_for_uuid(self, uuid: UUID) -> Path:
        """
        Generate a path in the temporary directory for a given UUID.

        This is the recommended way to store temporary data for the ingestion process. Note that this only returns the
        path, it does not create the directory or file.

        :param uuid: A UUID.
        :return: A Path object representing the path for the given UUID.
        """
        temp_dir = tempfile.gettempdir()
        return Path(os.path.join(temp_dir, f"{uuid}"))



import os
import pickle
import shutil
from dataclasses import dataclass
from datetime import timedelta, datetime
from enum import Enum
from pathlib import Path
from typing import Dict, Tuple, Callable
from uuid import UUID, uuid4

import sqlalchemy
from eflips.model.depot import Depot, AssocPlanProcess, Process, Area, Plan, AreaType
from eflips.model.general import Scenario, VehicleType
from eflips.model.network import Station, Line, Route, VoltageLevel, ChargeType
from eflips.model.schedule import Trip, Rotation, TripType
from sqlalchemy import create_engine
from sqlalchemy.orm import Session

from eflips.ingest.base import AbstractIngester


class BusType(Enum):
    """
    An enumeration of bus types.
    """

    SINGLE_DECKER = "Single Decker"
    DOUBLE_DECKER = "Double Decker"
    ARTICULATED = "Articulated"


@dataclass
class PrepareOptions:
    name: str
    depot_count: int
    line_count: int
    rotation_per_line: int
    opportunity_charging: bool
    bus_type: BusType


class DummyIngester(AbstractIngester):
    """
    The dummy ingester is a dummy implementation of the :class:`IngestBase` class. It is used
    to demonstrate the implementation.
    """

    @staticmethod
    def _create_vehicle_types(scenario: Scenario, session: Session, bus_type: BusType) -> None:
        """
        Create the vehicle types in the database.

        :param session: The database session.
        :return: None
        """
        # Add a vehicle type without a battery type
        # Doppeldecker mit mittlerer Batterie

        match bus_type:
            case BusType.SINGLE_DECKER:
                vehicle_type_1 = VehicleType(
                    scenario=scenario,
                    name="Bus Typ Dagobert",
                    battery_capacity=200,
                    charging_curve=[[0, 200], [1, 150]],
                    opportunity_charging_capable=True,
                    consumption=1,
                )
                session.add(vehicle_type_1)
            case BusType.DOUBLE_DECKER:
                # Add second vehicle type without a battery type
                # Kleiner Bus mit kleiner Batterie
                vehicle_type_2 = VehicleType(
                    scenario=scenario,
                    name="Bus Typ Düsentrieb",
                    battery_capacity=100,
                    charging_curve=[[0, 150], [1, 150]],
                    opportunity_charging_capable=True,
                    consumption=1,
                )
                session.add(vehicle_type_2)
            case BusType.ARTICULATED:
                # Add third vehicle type without a battery type
                # Langer Bus mit großer Batterie
                vehicle_type_3 = VehicleType(
                    scenario=scenario,
                    name="Bus Typ Panzerknacker",
                    battery_capacity=300,
                    charging_curve=[[0, 450], [1, 350]],
                    opportunity_charging_capable=True,
                    consumption=1,
                )
                session.add(vehicle_type_3)
            case _:
                raise ValueError("Invalid bus type")

    def prepare(  # type: ignore[override]
        self,
        name: str,
        depot_count: int,
        line_count: int,
        rotation_per_line: int,
        opportunity_charging: bool,
        bus_type: BusType,
        random_text_file: Path,
        progress_callback: None | Callable[[float], None] = None,
    ) -> Tuple[bool, UUID | Dict[str, str]]:
        """
        Dummy prepare method.

        It just checks the values, and then dumps them into a pickle in the temporary directory.

        :param name: The name of the scenario.
        :param depot_count: How many depots to create.
        :param line_count: How many lines to create.
        :param rotation_per_line: How many rotations to create per line.
        :param opportunity_charging: Whether opportunity charging is enabled.
        :param bus_type: Which bus type to use.
        :return: Either a UUID or a dictionary with the error message.
        """

        errors = {}

        if not isinstance(name, str) or len(name) == 0:
            errors["Wrong Name"] = "Name must be a non-empty string."
        if not isinstance(depot_count, int) or depot_count < 1:
            errors["Wrong Depot Count"] = "Depot count must be a positive integer."
        if not isinstance(line_count, int) or line_count < 1:
            errors["Wrong Line Count"] = "Line count must be a positive integer."
        if not isinstance(rotation_per_line, int) or rotation_per_line < 1:
            errors["Wrong Rotation Per Line"] = "Rotation per line must be a positive integer."
        if not isinstance(opportunity_charging, bool):
            errors["Wrong Opportunity Charging"] = "Opportunity charging must be a boolean."
        if not isinstance(bus_type, BusType):
            errors["Wrong Bus Type"] = "Bus type must be a valid bus type."
        if not random_text_file.name.endswith(".txt"):
            errors["Wrong Random Text File"] = "Random text file must end in .txt."

        if errors:
            return False, errors
        else:
            uuid = uuid4()
            temp_dir = self.path_for_uuid(uuid)
            temp_dir.mkdir(parents=True, exist_ok=False)
            shutil.move(random_text_file, temp_dir / random_text_file.name)

            data = PrepareOptions(
                name=name,
                depot_count=depot_count,
                line_count=line_count,
                rotation_per_line=rotation_per_line,
                opportunity_charging=opportunity_charging,
                bus_type=bus_type,
            )

            with open(temp_dir / "data.pkl", "wb") as f:
                pickle.dump(data, f)

            if progress_callback:
                progress_callback(1.0)

            return True, uuid

    def ingest(self, uuid: UUID, progress_callback: None | Callable[[float], None] = None) -> None:
        if not os.path.exists(self.path_for_uuid(uuid) / "data.pkl"):
            raise ValueError("Data file does not exist.")

        # Also check that there is one file ending in .txt
        if not any(p.name.endswith(".txt") for p in self.path_for_uuid(uuid).iterdir()):
            raise ValueError("No text file found.")

        with open(self.path_for_uuid(uuid) / "data.pkl", "rb") as f:
            params = pickle.load(f)
        assert isinstance(params, PrepareOptions)

        engine = create_engine(self.database_url)

        with Session(engine) as session:
            scenario_or_none = session.query(Scenario).filter(Scenario.task_id == uuid).one_or_none()
            if scenario_or_none:
                scenario = scenario_or_none
            else:
                scenario = Scenario(
                    name=params.name,
                )
                session.add(scenario)

            self._create_vehicle_types(scenario, session, params.bus_type)
            session.flush()  # To put the vehicle types into the database

            for i in range(params.depot_count):
                depot = self._create_depot(scenario, session, i)
                for j in range(params.line_count):
                    line = self._create_line(scenario, session, depot, i, j, params.opportunity_charging)
                    for k in range(params.rotation_per_line):
                        self._create_rotation(scenario, session, line, k, params.opportunity_charging)
                        if progress_callback:
                            progress_callback(
                                (i * params.line_count * params.rotation_per_line + j * params.rotation_per_line + k)
                                / (params.depot_count * params.line_count * params.rotation_per_line)
                            )

            session.commit()

    @classmethod
    def prepare_param_names(self) -> Dict[str, str | Dict[Enum, str]]:
        return {
            "name": "Name",
            "depot_count": "Depot Count",
            "line_count": "Lines per Depot",
            "rotation_per_line": "Rotations per Line",
            "opportunity_charging": "Opportunity Charging",
            "bus_type": {
                BusType.SINGLE_DECKER: "Single Decker",
                BusType.DOUBLE_DECKER: "Double Decker",
                BusType.ARTICULATED: "Articulated",
            },
            "random_text_file": "Random Text File",
        }

    @classmethod
    def prepare_param_description(self) -> Dict[str, str | Dict[Enum, str]]:
        return {
            "name": "The name of the scenario.",
            "depot_count": "The number of depots to create.",
            "line_count": "The number of lines to create. For each line, two routes are created.",
            "rotation_per_line": "The number of rotations per line. For each rotation, 20 trips are created.",
            "opportunity_charging": "Whether opportunity charging is enabled. If true, one of the terminals will have a charging station.",
            "bus_type": {
                BusType.SINGLE_DECKER: "Single Decker",
                BusType.DOUBLE_DECKER: "Double Decker",
                BusType.ARTICULATED: "Articulated",
            },
            "random_text_file": "A random text file that is not used for anything. Must end in .txt.",
        }

    def _create_depot(self, scenario: Scenario, session: sqlalchemy.orm.Session, i: int) -> Depot:
        """
        Creates a dummy depot.

        :param scenario: The scenario
        :param session: An SQLAlchemy session
        :param i: The number of the depot
        :return: The depot object (it is already added to the session)
        """

        station = Station(scenario=scenario, name=f"Station for Depot {i}", is_electrified=False)
        session.add(station)

        depot = Depot(scenario=scenario, name=f"Depot {i}", name_short=f"D{i}", station=station)
        session.add(depot)

        # Dirtily load the vehicle type
        with session.no_autoflush:
            vehicle_type = session.query(VehicleType).filter(VehicleType.scenario_id == scenario.id).one()

        # Create plan
        plan = Plan(scenario=scenario, name="Entenhausen Plan")
        session.add(plan)

        depot.default_plan = plan

        # Create areas
        arrival_area = Area(
            scenario=scenario,
            name="Entenhausen Depot Arrival Area",
            depot=depot,
            area_type=AreaType.DIRECT_ONESIDE,
            capacity=6,
            vehicle_type=vehicle_type,
        )
        session.add(arrival_area)

        cleaning_area = Area(
            scenario=scenario,
            name="Entenhausen Depot Cleaning Area",
            depot=depot,
            area_type=AreaType.DIRECT_ONESIDE,
            capacity=6,
            vehicle_type=vehicle_type,
        )
        session.add(cleaning_area)

        charging_area = Area(
            scenario=scenario,
            name="Entenhausen Depot Area",
            depot=depot,
            area_type=AreaType.LINE,
            row_count=2,
            capacity=6,
            vehicle_type=vehicle_type,
        )
        session.add(charging_area)

        # Create processes
        standby_arrival = Process(
            name="Standby Arrival",
            scenario=scenario,
            dispatchable=False,
        )
        session.add(standby_arrival)

        clean = Process(
            name="Clean",
            scenario=scenario,
            dispatchable=False,
            duration=timedelta(minutes=30),
        )
        session.add(clean)

        charging = Process(
            name="Charging",
            scenario=scenario,
            dispatchable=False,
            electric_power=150,
        )
        session.add(charging)

        standby_departure = Process(
            name="Standby Departure",
            scenario=scenario,
            dispatchable=True,
        )
        session.add(standby_departure)

        # Connect the areas and processes. *The final area needs to have both a charging and standby_departure process*

        arrival_area.processes.append(standby_arrival)
        cleaning_area.processes.append(clean)
        charging_area.processes.append(charging)
        charging_area.processes.append(standby_departure)

        assocs = [
            AssocPlanProcess(scenario=scenario, process=standby_arrival, plan=plan, ordinal=0),
            AssocPlanProcess(scenario=scenario, process=clean, plan=plan, ordinal=1),
            AssocPlanProcess(scenario=scenario, process=charging, plan=plan, ordinal=2),
            AssocPlanProcess(scenario=scenario, process=standby_departure, plan=plan, ordinal=3),
        ]
        session.add_all(assocs)

        return depot

    def _create_line(
        self,
        scenario: Scenario,
        session: sqlalchemy.orm.Session,
        depot: Depot,
        i: int,
        j: int,
        opportuinity_charging: bool,
    ) -> Line:
        bus_line = Line(scenario=scenario, name=f"Bus Line {i}-{j}")
        session.add(bus_line)

        # Create an outbound terminal
        outbound_terminal = Station(scenario=scenario, name=f"Outbound Terminal for {i}-{j}")
        if opportuinity_charging:
            outbound_terminal.is_electrified = True
            outbound_terminal.amount_charging_places = 1
            outbound_terminal.power_per_charger = 150
            outbound_terminal.power_total = 150
            outbound_terminal.voltage_level = VoltageLevel.MV
            outbound_terminal.charge_type = ChargeType.OPPORTUNITY
        else:
            outbound_terminal.is_electrified = False
        session.add(outbound_terminal)

        # Create outbound and inbound routes
        outbound_route = Route(
            scenario=scenario,
            name=f"Outbound Route for {i}-{j}",
            line=bus_line,
            departure_station=depot.station,
            arrival_station=outbound_terminal,
            distance=5000,
        )
        session.add(outbound_route)
        inbound_route = Route(
            scenario=scenario,
            name=f"Inbound Route for {i}-{j}",
            line=bus_line,
            departure_station=outbound_terminal,
            arrival_station=depot.station,
            distance=5000,
        )
        session.add(inbound_route)

        return bus_line

    def _create_rotation(
        self, scenario: Scenario, session: sqlalchemy.orm.Session, line: Line, k: int, opportunity_charging: bool
    ) -> None:
        # dirtily load the vehicle type
        vehicle_type = session.query(VehicleType).filter(VehicleType.scenario_id == scenario.id).one()

        # Create 10 inbound and outbound trips with 20 minutes between them offset by k*10 minutes
        first_departure = datetime(2024, 1, 1, 6, 0) + timedelta(minutes=k * 10)
        trip_duration = timedelta(minutes=30)
        break_duration = timedelta(minutes=10)

        inbound_route = session.query(Route).filter(Route.line_id == line.id, Route.name.like("Inbound%")).one()
        outbound_route = session.query(Route).filter(Route.line_id == line.id, Route.name.like("Outbound%")).one()

        next_departure = first_departure

        rotation = Rotation(
            scenario=scenario,
            name=f"Rotation {line.name} {k}",
            vehicle_type=vehicle_type,
            allow_opportunity_charging=opportunity_charging,
        )
        session.add(rotation)

        for i in range(10):
            outbound_trip = Trip(
                scenario=scenario,
                rotation=rotation,
                route=outbound_route,
                departure_time=next_departure,
                arrival_time=next_departure + trip_duration,
                trip_type=TripType.PASSENGER,
            )
            session.add(outbound_trip)
            next_departure += trip_duration + break_duration

            inbound_trip = Trip(
                scenario=scenario,
                rotation=rotation,
                route=inbound_route,
                departure_time=next_departure,
                arrival_time=next_departure + trip_duration,
                trip_type=TripType.PASSENGER,
            )
            session.add(inbound_trip)
            next_departure += trip_duration + break_duration



"""This file was generated by xsdata, v24.1, on 2024-04-29 12:10:47

Generator: DataclassGenerator
See: https://xsdata.readthedocs.io/
"""
from eflips.ingest.vdv452data.vdv452_v1_5 import (
    BasisVerGueltigkeit as BasisVerGueltigkeit,
    Firmenkalender as Firmenkalender,
    LidVerlauf as LidVerlauf,
    MengeFzgTyp as MengeFzgTyp,
    OrtHztf as OrtHztf,
    RecFrt as RecFrt,
    RecFrtHzt as RecFrtHzt,
    RecLid as RecLid,
    RecOrt as RecOrt,
    RecSel as RecSel,
    RecUeb as RecUeb,
    RecUmlauf as RecUmlauf,
    SelFztFeld as SelFztFeld,
    UebFzt as UebFzt,
    VdvBaseObject as VdvBaseObject,
)



from typing import List, Type

from eflips.ingest.base import AbstractIngester
from eflips.ingest.dummy import DummyIngester


def all_ingesters() -> List[Type[AbstractIngester]]:
    return [DummyIngester]



import os

from eflips.ingest.vdv import validate_input_data_vdv_451, VdvRequiredTables, import_vdv452_table_records

if __name__ == "__main__":
    path_to_this_file = os.path.dirname(os.path.abspath(__file__))
    sample_files_dir = os.path.join(path_to_this_file, "..", "..", "UVG")
    all_tables = validate_input_data_vdv_451(sample_files_dir)

    all_data = {}

    for tbl in all_tables:
        if tbl in VdvRequiredTables.required_tables.keys():
            all_data[tbl] = import_vdv452_table_records(all_tables[tbl])

    print("Done.")



import csv
import enum
import glob
import logging
import os
import pickle
import re
from dataclasses import dataclass
from datetime import date, timedelta, datetime, time
from enum import Enum
from pathlib import Path
from typing import Dict, Callable, Tuple, Optional, List
from uuid import UUID, uuid4
from zipfile import ZipFile

import pytz
from eflips.model import VehicleType, Scenario, Rotation, Station, Line, Route, Trip, TripType, StopTime
from sqlalchemy import create_engine
from sqlalchemy.orm import Session
from tqdm.auto import tqdm

from eflips.ingest.base import AbstractIngester
from eflips.ingest.vdv452data import (
    VdvBaseObject,
    BasisVerGueltigkeit,
    Firmenkalender,
    RecOrt,
    MengeFzgTyp,
    RecSel,
    SelFztFeld,
    LidVerlauf,
    RecFrt,
    RecUmlauf,
    RecLid,
    RecFrtHzt,
    OrtHztf,
)


class VDV_Data_Type(enum.Enum):
    """
    An enum for the data types as specified in VDV 451/452. We map the different data types to two main types:
    - CHAR: Character data
    - NUM: Numeric data
    """

    CHAR = "char"
    INT = "num"
    FLOAT = enum.auto


class VDV_Table_Name(enum.Enum):
    """
    An enum for the table names as Specified in VDV 451/452. Only the strings are used here as the numbers
    have never been encountered in practice.
    """

    MENGE_BASIS_VERSIONEN = "MENGE_BASIS_VERSIONEN"
    BASIS_VER_GUELTIGKEIT = "BASIS_VER_GUELTIGKEIT"
    FIRMENKALENDER = "FIRMENKALENDER"
    MENGE_TAGESART = "MENGE_TAGESART"
    MENGE_ONR_TYP = "MENGE_ONR_TYP"
    MENGE_ORT_TYP = "MENGE_ORT_TYP"
    REC_HP = "REC_HP"
    REC_OM = "REC_OM"
    REC_ORT = "REC_ORT"
    FAHRZEUG = "FAHRZEUG"
    ZUL_VERKEHRSBETRIEB = "ZUL_VERKEHRSBETRIEB"
    MENGE_BEREICH = "MENGE_BEREICH"
    MENGE_FZG_TYP = "MENGE_FZG_TYP"
    REC_ANR = "REC_ANR"
    REC_ZNR = "REC_ZNR"
    REC_SEL = "REC_SEL"
    REC_SEL_ZP = "REC_SEL_ZP"
    MENGE_FGR = "MENGE_FGR"
    ORT_HZTF = "ORT_HZTF"
    SEL_FZT_FELD = "SEL_FZT_FELD"
    REC_UEB = "REC_UEB"
    UEB_FZT = "UEB_FZT"
    MENGE_FAHRTART = "MENGE_FAHRTART"
    LID_VERLAUF = "LID_VERLAUF"
    REC_LID = "REC_LID"
    REC_FRT = "REC_FRT"
    REC_FRT_HZT = "REC_FRT_HZT"
    REC_UMLAUF = "REC_UMLAUF"


@dataclass
class VdvRequiredTables:
    """
    This class serves to list the required tables for the VDV 452 data ingestion.
    """

    # Required Tables:
    # BASIS_VER_GUELTIGKEIT: Nötig, um herauszufinden welches die aktuell gültige Version ist (kein Handling, falls nicht existent und trz nur 1 Version existiert)
    # FIRMENKALENDER: Brauchen wir wegen zuordnung Betriebstage zu Tagesarten (für Umläufe)
    # REC_ORT: sprechende Namen der Haltestellen, aber auch (optional) Längen- und Breitengrade, Höhe
    # MENGE_FZG_TYP: Angaben der Fahrzeugtypen (Länge, Breite, ..., u.U. Batterieinfos, Name, Verbräuche, ..)
    # REC_SEL: Nötig, enthalten Angaben über die km-Entfernung zwischen den Haltestellen
    # SEL_FZT_FELD: Fahrzeiten zwischen zwei Haltestellen (je Fahrzeitgruppe), definitiv nötig
    # LID_VERLAUF: Für die Linienverläufe (Stationen und Reihenfolge)
    # REC_FRT: brauchen wir für die Zuordnung Fahrt->Umlauf. In der Tabelle findet die Zuordnung der Fahrt zur Linie und Tagesart.
    # REC_UMLAUF: Hier sind die eigentlichen Umläufe beschrieben

    # nicht ganz klar / Kontextabhängig:
    # Haltezeiten: ggfs. entweder-oder aus:
    #   a) ORT_HZTF: Haltezeiten je Fahrzeitgruppe und Ort/Haltestelle.
    #   b) REC_FRT_HZT: Haltezeiten je Fahrt. Also spezifischer als ORT_HZTF. Mir ist unklar, was passiert, falls beide vorhanden sind. Finde ich nichts in der Doku. Würde dann immer das hier genommen?
    #   c) oder auf 0 falls nicht existent?
    # Überläuferfahrten (VDV 452 Kap. 9.8.2 bzw. S. 65 bei "Erläuterung zur Fahrzeugumlaufbildung aus den Fahrten", demnach ist es kontextabhängig ob die folgenden Tabellen nötig sind):
    #   a) REC_UEB: Länge der Fahrt zwischen zwei Orten - bei Überläuferfahrt (Betriebshofaus- und -einfahrt, Zufahrten)
    #   b) UEB_FZT: Fahrzeiten bei Überläuferfahrten
    # !!! Anmerkung: Überläuferfahrten sind eben nicht immer so definiert, manchmal wird es wie eine "Fahrgastfahrt" behandelt, das schwankt etwas zwischen den mir vorliegenden VDV Datensätzen.
    # REC_LID: Linienvarianten. Ich würde sagen schon nötig, wegen der Linienbezeichnung; aber theoretisch könnte es ggfs. weggelassen werden (?) und mit dem PK gearbeitet werden
    # EINZELANSCHLUSS: für Anschlussdefinition, glaube mir machen keine Umstiege in eflips?
    # REC_UMS: Umsteigezeiten für Anschlusssicherung.

    # Später behandeln: 14 E-Mobilitäts Tabellen aus der VDV (Kapitel 11.6 - 11.14)

    # egal:
    # MENGE_BASIS_VERSIONEN: eigentlich nur für Textuelle Beschreibung der Basisversion nötig
    # MENGE_TAGESART: nur textuelle Beschreibung der Tagesart
    # MENGE_ONR_TYP: nur textuelle Beschreibung der funktionalen Ortstypen
    # MENGE_ORT_TYP: analog nur textuelle Beschreibung der Ortstypen
    # REC_HP: nur textuelle Beschreibung / Nummerierung der Haltepunkttypen
    # die beschreibung des Haltepunktes ist tatsächlich eig. egal, weil wir mit den REC_ORten arbeiten und nicht mit Haltepunkten
    # REC_OM: egal, weil nur nötig für Fahrzeug-Standortverfolgung im ITCS System (Ortsmarken)
    # FAHRZEUG: Machen wir im eflips-depot eh neu
    # ZUL_VERKEHRSBETRIEB: wir differenzieren nicht zwischen verschiedenen Verkehrsunternehmen, sondern simulieren alles
    # MENGE_BEREICH: wieder nur textuelle Beschreibung der Linienbereiche / Betriebszweige
    # REC_ANR: Ansagetexte sind uns egal
    # REC_ZNR: Angezeigte Fahrtziele, egal
    # REC_SEL_ZP: Geografischer Verlauf der Fahrt zwischen zwei Stationen - das betrachten wir nicht
    # MENGE_FGR: Textuelle Beschreibung von Fahrzeitgruppen, irrelevant
    # FLAECHEN_ZONE: Flächenzonen beschrieben, sollte uns egal sein
    # FL_ZONE_ORT: analog
    # MENGE_FLAECHEN_ZONE_TYP: analog
    # SEL_FZT_FELD_ZP: Fahrzeit für Zwischenpunkte. SOLLTE egal sein. todo kann es möglich sein, dass nur diese Zwischenpunkte angegeben sind oder so?!?!
    # MENGE_FAHRTART: Textuelle Beschreibung der Fahrtarten (z.B. Normalfahrt, Betriebshofausfahrt usw. siehe VDV 452 Kap. 9.6.8)

    # fahrzeug waere optional, aber machen wir nicht?

    # required tables: die Enum Datatypes als keys und das Dataclass object als value.
    required_tables = {
        VDV_Table_Name.BASIS_VER_GUELTIGKEIT: BasisVerGueltigkeit,
        VDV_Table_Name.FIRMENKALENDER: Firmenkalender,
        VDV_Table_Name.REC_ORT: RecOrt,
        VDV_Table_Name.MENGE_FZG_TYP: MengeFzgTyp,
        VDV_Table_Name.REC_SEL: RecSel,
        VDV_Table_Name.SEL_FZT_FELD: SelFztFeld,
        VDV_Table_Name.LID_VERLAUF: LidVerlauf,
        VDV_Table_Name.REC_FRT: RecFrt,
        VDV_Table_Name.REC_UMLAUF: RecUmlauf,
        VDV_Table_Name.REC_LID: RecLid,
    }


@dataclass
class VDVTable:
    abs_file_path: str | Path
    character_set: str
    table_name: VDV_Table_Name
    column_names_and_data_types: list[
        Tuple[str, Optional[VDV_Data_Type]]
    ]  # None (optional) in the VDV_Data_Type represents "other / invalid data type" here


def fix_identical_stop_times(stop_times: List[StopTime]) -> None:
    """
    This function goes through a list of stop times and changes the arrival time of a stop time to be the same as the

    :param stop_times: A list of stop times. The list is assumed to be sorted by arrival time.
    :return: Nothing. The list is modified in place.
    """
    # First, identify the indizes of the stop times that have the same arrival time
    indizes_of_identical_arrival_times: Dict[datetime, List[int]] = {}
    for i, stop_time in enumerate(stop_times):
        if stop_time.arrival_time not in indizes_of_identical_arrival_times:
            indizes_of_identical_arrival_times[stop_time.arrival_time] = []
        indizes_of_identical_arrival_times[stop_time.arrival_time].append(i)
    indizes_of_identical_arrival_times = {k: v for k, v in indizes_of_identical_arrival_times.items() if len(v) > 1}

    # Now depending on the length of the list, we have to adjust the arrival times, so they are evenly spaced
    # throughout a minute (e.g. with 2 stops, the first one arrives at 12:00:00 and the second one at 12:00:30)
    # We do this by adding a timedelta to the arrival time of the stop time.
    # However, if the stop time is the last one in the list, we then need to *subtract* the offest, so the last
    # time stays the same

    for identical_arrival_times in indizes_of_identical_arrival_times.values():
        assert len(identical_arrival_times) > 1
        # We cannot assume the minimum resolution is one minute. So we need to check the minimum difference
        # Before and after
        if identical_arrival_times[0] != 0:
            diff_before = (
                stop_times[identical_arrival_times[0]].arrival_time
                - stop_times[identical_arrival_times[0] - 1].arrival_time
            )
        else:
            diff_before = timedelta(seconds=60)
        if identical_arrival_times[-1] != len(stop_times) - 1:
            diff_after = (
                stop_times[identical_arrival_times[-1] + 1].arrival_time
                - stop_times[identical_arrival_times[-1]].arrival_time
            )
        else:
            diff_after = timedelta(seconds=60)
        # We take the minimum of the two
        offset = min(diff_before, diff_after) / len(identical_arrival_times)
        for i, idx in enumerate(identical_arrival_times):
            stop_times[idx].arrival_time += i * offset
        if idx == len(stop_times) - 1:
            # This is how much we shifted the last stop time, so we need to subtract it again
            max_offset = (len(identical_arrival_times) - 1) * offset
            for idx in identical_arrival_times:
                stop_times[idx].arrival_time -= max_offset


class VdvIngester(AbstractIngester):
    def prepare(  # type: ignore[override]
        self,
        x10_zip_file: Path,
        progress_callback: None | Callable[[float], None] = None,
    ) -> Tuple[bool, UUID | Dict[str, str]]:
        """
        Prepare and validate the input data for ingestion.

        :param x10_zip_file: A pathlib.Path object representing the path to the x10 zip file.
        :param progress_callback: A function that accepts a float value between 0 and 1. This function will be called
                                  periodically to update the progress of the ingestion process.
        :return: A tuple containing a boolean indicating whether the input data is valid and either a UUID or a dictionary
                 containing the error message.
        """

        # Check that the zip file only contains x10 files and that they are not empty or too large
        valid_or_error = validate_zip_file(x10_zip_file)
        if valid_or_error is not True:
            assert isinstance(valid_or_error, dict)
            return False, valid_or_error

        # Generate a uuid and extract the zip file to the temporary directory
        uuid = uuid4()
        dir = self.path_for_uuid(uuid)
        os.makedirs(dir, exist_ok=False)
        with ZipFile(x10_zip_file, "r") as zip_file:
            zip_file.extractall(dir)

        # Check if all the required tables are present in the directory
        try:
            all_tables = validate_input_data_vdv_451(dir)
        except ValueError as e:
            return False, {"validation": str(e)}

        with open(dir / "all_tables.pkl", "wb") as fp:
            pickle.dump(all_tables, fp)

        # If all tables are present, return the UUID
        return True, uuid

    def ingest(self, uuid: UUID, progress_callback: None | Callable[[float], None] = None) -> None:
        logger = logging.getLogger(__name__)

        # Load the paths to the tables
        temp_dir = self.path_for_uuid(uuid)
        all_tables_file = Path(temp_dir) / "all_tables.pkl"
        with open(all_tables_file, "rb") as fp:
            all_tables = pickle.load(fp)

        # For each table, turn it into a list of VDV base objects
        all_data: Dict[VDV_Table_Name, List[VdvBaseObject]] = {}
        for tbl in all_tables:
            all_data[tbl] = import_vdv452_table_records(all_tables[tbl])

        # Now, we have all the data in the all_data dictionary. For each data piece,
        # - put it in the database in the correct object
        # - if we need to reference it later, put it into a dictionary, where we store the eflips-model object against
        #   its VDV-style primary key

        engine = create_engine(self.database_url)
        with Session(engine) as session:
            try:
                scenario = Scenario(name=f"Created from VDV data with UUID {uuid}")
                session.add(scenario)

                # Vehicle Types
                vehicle_types_by_vdv_pk: Dict[Tuple[int | date | str, ...], VehicleType] = {}
                for vdv_vehicle_type in all_data[VDV_Table_Name.MENGE_FZG_TYP]:
                    assert isinstance(vdv_vehicle_type, MengeFzgTyp)
                    db_vehicle_type = vdv_vehicle_type.to_vehicle_type(scenario)
                    session.add(db_vehicle_type)
                    vehicle_types_by_vdv_pk[vdv_vehicle_type.primary_key] = db_vehicle_type

                # Rotations
                rotations_by_vdv_pk: Dict[Tuple[int | str | date, ...], Rotation] = {}

                # We may need a dummy vehicle type for rotations that do not have a vehicle type associated with them
                dummy_vehicle_type: VehicleType | None = None

                for vdv_rotation in all_data[VDV_Table_Name.REC_UMLAUF]:
                    assert isinstance(vdv_rotation, RecUmlauf)

                    if vdv_rotation.fzg_typ_nr is None and dummy_vehicle_type is None:
                        # We need to add a dummy vehicle type for the first time, create it
                        dummy_vehicle_type = self.create_dummy_vehicle_type(scenario)
                        session.add(dummy_vehicle_type)

                    db_rotation = vdv_rotation.to_rotation(
                        scenario, vehicle_types_by_vdv_pk, dummy_vehicle_type=dummy_vehicle_type
                    )
                    session.add(db_rotation)
                    rotations_by_vdv_pk[vdv_rotation.primary_key] = db_rotation

                # Stations

                # Stations we handle specially. Since the VDV RecOrt is more what we would call the
                # "coordinates of RouteStopAssociations" in eflips-model, we use a method there to extract the actual
                # Station object from it.
                assert all(isinstance(x, RecOrt) for x in all_data[VDV_Table_Name.REC_ORT])
                rec_orts: List[RecOrt] = [x for x in all_data[VDV_Table_Name.REC_ORT] if isinstance(x, RecOrt)]
                stations_by_vdv_pk: Dict[Tuple[int | date | str, ...], Station] = RecOrt.list_of_stations(
                    rec_orts, scenario
                )
                # The values of this dict are not unique, so we only add the unique ones to the database
                for station in set(stations_by_vdv_pk.values()):
                    session.add(station)

                # Lines
                # The same Line might be shared by multiple RecLid objects, but is unique by li_kuerzel
                lines_by_li_kuerzel: Dict[str, Line] = {}
                lines_by_vdv_pk: Dict[Tuple[int | date | str, ...], Line] = {}

                assert all(isinstance(x, RecLid) for x in all_data[VDV_Table_Name.REC_LID])
                rec_lids = [x for x in all_data[VDV_Table_Name.REC_LID] if isinstance(x, RecLid)]

                for rec_lid in rec_lids:
                    line_name = rec_lid.li_kuerzel
                    if line_name not in lines_by_li_kuerzel:
                        line = Line(name=line_name, scenario=scenario)
                        session.add(line)
                        lines_by_li_kuerzel[line_name] = line
                    else:
                        line = lines_by_li_kuerzel[line_name]
                    lines_by_vdv_pk[rec_lid.primary_key] = line

                # Routes
                # Those are more intricate, as we will have to compose them from the RecLid and RecSel and ??? tables
                # We need to construct quite a few helper dictionaries for this
                lines_by_basis_version_and_li_nr: Dict[Tuple[int | date | str, ...], Line] = {
                    (k[0], k[1]): v for k, v in lines_by_vdv_pk.items()
                }

                rec_orts_by_basis_version_and_onr_typ_nr_and_ort_nr = {
                    (r.basis_version, r.onr_typ_nr, r.ort_nr): r for r in rec_orts
                }

                assert all(isinstance(x, RecSel) for x in all_data[VDV_Table_Name.REC_SEL])
                rec_sels: List[RecSel] = [x for x in all_data[VDV_Table_Name.REC_SEL] if isinstance(x, RecSel)]
                rec_sel_by_basis_version_and_start_type_and_start_nr_and_end_type_and_end_nr = {
                    (r.basis_version, r.onr_typ_nr, r.ort_nr, r.sel_ziel_typ, r.sel_ziel): r for r in rec_sels
                }

                assert all(isinstance(x, LidVerlauf) for x in all_data[VDV_Table_Name.LID_VERLAUF])
                lid_verlaufs: List[LidVerlauf] = [
                    x for x in all_data[VDV_Table_Name.LID_VERLAUF] if isinstance(x, LidVerlauf)
                ]
                lid_verlaufs_by_basis_version_and_li_nr_and_str_li_var: Dict[
                    Tuple[int, int, str], List[LidVerlauf]
                ] = {}
                for lid_verlauf in lid_verlaufs:
                    key = (lid_verlauf.basis_version, lid_verlauf.li_nr, lid_verlauf.str_li_var)
                    if key not in lid_verlaufs_by_basis_version_and_li_nr_and_str_li_var:
                        lid_verlaufs_by_basis_version_and_li_nr_and_str_li_var[key] = []
                    # Insert it in the correct order, by the lid_verlauf.li_lfd_nr
                    # Runtime expensive, but works
                    lid_verlaufs_by_basis_version_and_li_nr_and_str_li_var[key].append(lid_verlauf)
                    lid_verlaufs_by_basis_version_and_li_nr_and_str_li_var[key].sort(key=lambda x: x.li_lfd_nr)

                # Now we can construct the routes
                routes_by_vdv_pk: Dict[Tuple[int | date | str, ...], Route] = {}
                assert all(isinstance(x, RecLid) for x in all_data[VDV_Table_Name.REC_LID])
                rec_lids = [x for x in all_data[VDV_Table_Name.REC_LID] if isinstance(x, RecLid)]

                rec_selss: Dict[
                    Tuple[int | date | str, ...], List[RecSel]
                ] = {}  # Later we will use this to construct the trips

                for rec_lid in rec_lids:
                    route: Route
                    route, rec_selss[(rec_lid.basis_version, rec_lid.li_nr, rec_lid.str_li_var)] = rec_lid.to_route(
                        scenario=scenario,
                        lines_by_basis_version_andli_nr=lines_by_basis_version_and_li_nr,
                        rec_orts_by_basis_version_and_onr_typ_nr_and_ort_nr=rec_orts_by_basis_version_and_onr_typ_nr_and_ort_nr,
                        lid_verlaufs_by_basis_version_and_li_nr_and_str_li_var=lid_verlaufs_by_basis_version_and_li_nr_and_str_li_var,
                        stations_by_basis_version_and_onr_typ_nr_and_ort_nr=stations_by_vdv_pk,
                        rec_sel_by_basis_version_and_start_type_and_start_nr_and_end_type_and_end_nr=rec_sel_by_basis_version_and_start_type_and_start_nr_and_end_type_and_end_nr,
                    )
                    session.add(route)
                    routes_by_vdv_pk[rec_lid.primary_key] = route

                # Now we can construct the trips

                assert all(isinstance(x, SelFztFeld) for x in all_data[VDV_Table_Name.SEL_FZT_FELD])
                sel_fzt_felds = [x for x in all_data[VDV_Table_Name.SEL_FZT_FELD] if isinstance(x, SelFztFeld)]
                sel_fzt_felds_by_pk: Dict[Tuple[int | date | str, ...], List[SelFztFeld]] = {}
                for this_sel_fzt_feld in sel_fzt_felds:
                    pk = this_sel_fzt_feld.primary_key
                    if pk not in sel_fzt_felds_by_pk:
                        sel_fzt_felds_by_pk[pk] = []
                    sel_fzt_felds_by_pk[pk].append(this_sel_fzt_feld)

                # We also need to create a slighltly relaxes sel_fzt_felds_by_pk, where we only have one entry per pk
                sel_fzt_felds_by_relaxed_pk: Dict[Tuple[int | date | str, ...], List[SelFztFeld]] = {}
                for x in sel_fzt_felds:
                    relaxed_pk = (x.basis_version, x.bereich_nr, x.onr_typ_nr, x.ort_nr, x.sel_ziel_typ, x.sel_ziel)
                    if relaxed_pk not in sel_fzt_felds_by_relaxed_pk:
                        sel_fzt_felds_by_relaxed_pk[relaxed_pk] = []
                    sel_fzt_felds_by_relaxed_pk[relaxed_pk].append(x)

                assert all(isinstance(x, RecFrt) for x in all_data[VDV_Table_Name.REC_FRT])
                rec_frts = [x for x in all_data[VDV_Table_Name.REC_FRT] if isinstance(x, RecFrt)]

                if VDV_Table_Name.ORT_HZTF in all_data.keys():
                    assert all(isinstance(x, OrtHztf) for x in all_data[VDV_Table_Name.ORT_HZTF])
                    ort_hztfs = [x for x in all_data[VDV_Table_Name.ORT_HZTF] if isinstance(x, OrtHztf)]
                else:
                    ort_hztfs = []

                if VDV_Table_Name.REC_FRT_HZT in all_data.keys():
                    assert all(isinstance(x, RecFrtHzt) for x in all_data[VDV_Table_Name.REC_FRT_HZT])
                    rec_frt_hzts = [x for x in all_data[VDV_Table_Name.REC_FRT_HZT] if isinstance(x, RecFrtHzt)]
                else:
                    rec_frt_hzts = []

                # It only makes sense to have one of the two
                assert bool(len(rec_frt_hzts) > 0) ^ bool(len(ort_hztfs) > 0)

                assert all(isinstance(x, Firmenkalender) for x in all_data[VDV_Table_Name.FIRMENKALENDER])
                firmenkalenders = [x for x in all_data[VDV_Table_Name.FIRMENKALENDER] if isinstance(x, Firmenkalender)]

                rotations_by_vdv_pk_and_date: Dict[Tuple[int, int, int, date], Rotation] = dict()

                for rec_frt in tqdm(rec_frts):
                    # Load the dwell durations rec-frt-hzt that belong to this trip
                    this_trip_rec_frt_hzts = [x for x in rec_frt_hzts if x.frt_fid == rec_frt.frt_fid]

                    # Load the corresponding Route object
                    route = routes_by_vdv_pk[(rec_frt.basis_version, rec_frt.li_nr, rec_frt.str_li_var)]

                    # Also, load the corresponding rec_sels
                    this_route_rec_sels: List[RecSel] = rec_selss[
                        (rec_frt.basis_version, rec_frt.li_nr, rec_frt.str_li_var)
                    ]
                    this_route_sel_fzt_felds = []
                    for rec_sel in this_route_rec_sels:
                        sel_fzt_feld_pk = (
                            rec_sel.basis_version,
                            rec_sel.bereich_nr,
                            rec_frt.fgr_nr,
                            rec_sel.onr_typ_nr,
                            rec_sel.ort_nr,
                            rec_sel.sel_ziel_typ,
                            rec_sel.sel_ziel,
                        )

                        if sel_fzt_feld_pk in sel_fzt_felds_by_pk.keys():
                            sel_fzt_feld = sel_fzt_felds_by_pk[sel_fzt_feld_pk]
                        else:
                            logger.debug(f"Could not find SelFztFeld for {sel_fzt_feld_pk}")
                            # Find one by relaxing the constraints
                            sel_fzt_feld = sel_fzt_felds_by_relaxed_pk[
                                (
                                    rec_sel.basis_version,
                                    rec_sel.bereich_nr,
                                    rec_sel.onr_typ_nr,
                                    rec_sel.ort_nr,
                                    rec_sel.sel_ziel_typ,
                                    rec_sel.sel_ziel,
                                )
                            ]
                            sel_fzt_feld = [sel_fzt_feld[0]]

                        if len(sel_fzt_feld) != 1:
                            logger.info(f"Could not find exactly one SelFztFeld for {sel_fzt_feld_pk}")
                            # If there are more than one, make sure the durations are the same
                            if len(sel_fzt_feld) > 1:
                                durations = [x.sel_fzt for x in sel_fzt_feld]
                                if len(set(durations)) != 1:
                                    logger.warning(
                                        f"Multiple SelFztFelds for {sel_fzt_feld_pk} have different durations: {durations}"
                                    )
                                    raise ValueError(
                                        f"Multiple SelFztFelds for {sel_fzt_feld_pk} have different durations: {durations}"
                                    )
                                else:
                                    duration = durations[0]
                            else:
                                # Length is 0 -- create a zero duration
                                duration = timedelta(minutes=0)
                            # For now, create a dummy one
                            sel_fzt_feld = [
                                SelFztFeld(
                                    basis_version=rec_sel.basis_version,
                                    bereich_nr=rec_sel.bereich_nr,
                                    fgr_nr=rec_frt.fgr_nr,
                                    onr_typ_nr=rec_sel.onr_typ_nr,
                                    ort_nr=rec_sel.ort_nr,
                                    sel_ziel_typ=rec_sel.sel_ziel_typ,
                                    sel_ziel=rec_sel.sel_ziel,
                                    sel_fzt=duration,
                                )
                            ]
                        this_route_sel_fzt_felds.append(sel_fzt_feld[0])

                    # Calculate the dwell durations and driving durations for each station
                    elapsed_duration = rec_frt.frt_start
                    arrival_time_from_start: List[timedelta] = []
                    dwell_durations: List[timedelta] = []
                    for i in range(len(this_route_rec_sels)):
                        cur_rec_sel = this_route_rec_sels[i]
                        # Add the first station
                        if i == 0:
                            # Check the rec_frt_hzts for the first station
                            first_station_pk = (cur_rec_sel.basis_version, cur_rec_sel.onr_typ_nr, cur_rec_sel.ort_nr)
                            first_station_rec_frt_hzts = [
                                x for x in this_trip_rec_frt_hzts if x.position_key == first_station_pk
                            ]

                            # Check the ort_hztfs for the first station
                            first_station_ort_hztfs = [x for x in ort_hztfs if x.position_key == first_station_pk]

                            if len(first_station_rec_frt_hzts) == 1:
                                dwell_duration = first_station_rec_frt_hzts[0].frt_hzt_zeit
                            elif len(first_station_ort_hztfs) == 1:
                                dwell_duration = first_station_ort_hztfs[0].hp_hzt
                            else:
                                logger.debug(
                                    f"Could not find any dwell duration for the station {first_station_pk}. Adding 0s."
                                )
                                # For now, create a dummy one
                                dwell_duration = timedelta(seconds=0)

                            arrival_time_from_start.append(
                                elapsed_duration
                            )  # For the first station, the arrival time is the start of the trip
                            dwell_durations.append(dwell_duration)  # dwell duration for the first station
                            elapsed_duration += dwell_duration

                        # Now, always add the driving duration and the dwell duration at the destination of the rec_sel
                        # First, add the driving duration
                        cur_sel_fzt_feld = this_route_sel_fzt_felds[i]
                        driving_duration = cur_sel_fzt_feld.sel_fzt
                        elapsed_duration += driving_duration
                        arrival_time_from_start.append(elapsed_duration)

                        # Load the dwell duration for the destination station of this segment
                        next_station_pk = (cur_rec_sel.basis_version, cur_rec_sel.sel_ziel_typ, cur_rec_sel.sel_ziel)
                        next_station_rec_frt_hzts = [
                            x for x in this_trip_rec_frt_hzts if x.position_key == next_station_pk
                        ]

                        next_station_ort_hztfs = [x for x in ort_hztfs if x.position_key == next_station_pk]

                        if len(next_station_rec_frt_hzts) == 1:
                            dwell_duration = next_station_rec_frt_hzts[0].frt_hzt_zeit
                        elif len(next_station_ort_hztfs) == 1:
                            dwell_duration = next_station_ort_hztfs[0].hp_hzt
                        else:
                            logger.debug(
                                f"Could not find any dwell duration for the station {next_station_pk}. Adding 0s."
                            )
                            # For now, create a dummy one
                            dwell_duration = timedelta(seconds=0)

                        dwell_durations.append(dwell_duration)
                        elapsed_duration += dwell_duration

                    ### CREATE THE TRIP
                    # We need to do this on all days that have the same tagesart as the rec_frt
                    # We need to find the tagesart of the rec_frt
                    for firmenkalender in firmenkalenders:
                        if firmenkalender.tagesart_nr == rec_frt.tagesart_nr:
                            the_date = firmenkalender.betriebstag

                            # Check if a specific rotation for this day exists
                            vdv_pk_and_date = (rec_frt.basis_version, rec_frt.tagesart_nr, rec_frt.um_uid, the_date)
                            if vdv_pk_and_date in rotations_by_vdv_pk_and_date:
                                rotation = rotations_by_vdv_pk_and_date[vdv_pk_and_date]
                            else:
                                orig_rotation = rotations_by_vdv_pk[
                                    (rec_frt.basis_version, rec_frt.tagesart_nr, rec_frt.um_uid)
                                ]
                                rotation = Rotation(
                                    scenario=scenario,
                                    name=orig_rotation.name,
                                    vehicle_type=orig_rotation.vehicle_type,
                                    trips=[],
                                    allow_opportunity_charging=orig_rotation.allow_opportunity_charging,
                                )
                                rotations_by_vdv_pk_and_date[vdv_pk_and_date] = rotation
                                session.add(rotation)

                            # Create a local midnight datetime object in the "Europe/Berlin" timezone
                            tz = pytz.timezone("Europe/Berlin")
                            local_midnight = tz.localize(datetime.combine(the_date, time(0, 0)))

                            # Create the trip, if it is a valid trip
                            if rec_frt.frt_start.total_seconds() != elapsed_duration.total_seconds():
                                trip = Trip(
                                    scenario=scenario,
                                    route=route,
                                    departure_time=local_midnight + rec_frt.frt_start,
                                    arrival_time=local_midnight + elapsed_duration,
                                    trip_type=TripType.PASSENGER if rec_frt.fahrtart_nr == 1 else TripType.EMPTY,
                                )

                                # Create the stop times
                                stop_times = []
                                for i in range(len(route.assoc_route_stations)):
                                    station = route.assoc_route_stations[i].station
                                    arrival_time = local_midnight + arrival_time_from_start[i]
                                    dwell_duration = dwell_durations[i]
                                    stop_time = StopTime(
                                        scenario=scenario,
                                        trip=trip,
                                        station=station,
                                        arrival_time=arrival_time,
                                        dwell_duration=dwell_duration,
                                    )
                                    stop_times.append(stop_time)

                                # Fix identical stop times
                                fix_identical_stop_times(stop_times)

                                # Look up the rotation using the basis_version and um_uid
                                trip.rotation = rotation
                                session.add(trip)
                            else:
                                raise ValueError(f"Trip {rec_frt.frt_fid} has a duration of 0 seconds. Skipping.")

                # Delete all rotations in this scenario with no trips
                session.flush()
                for rotation in scenario.rotations:
                    if len(rotation.trips) == 0:
                        session.delete(rotation)

            except Exception as e:
                session.rollback()
                raise e
            finally:
                session.commit()

    @classmethod
    def create_dummy_vehicle_type(cls, scenario: Scenario) -> VehicleType:
        dummy_vehicle_type = VehicleType(
            scenario=scenario,
            name="Dummy Vehicle Type",
            opportunity_charging_capable=False,
            battery_capacity=10000,
            charging_curve=[[0, 1], [1000, 1000]],
        )
        return dummy_vehicle_type

    @classmethod
    def prepare_param_names(cls) -> Dict[str, str | Dict[Enum, str]]:
        """
        A dictionary containing the parameter names for :meth:`prepare`.

        These should be short, descriptive names for the parameters of the :meth:`prepare` method. The keys must be the
        names of the parameters, and the values should be strings describing the parameter. If the keyword argument is
        an enumerated type, the value should be a dictionary with the keys being the members.

        This method can then be used to generate a help text for the user.

        :return: A dictionary containing the parameter hints for the prepare method.
        """
        return {
            "x10_zip_file": "VDV data archive",
        }

    @classmethod
    def prepare_param_description(cls) -> Dict[str, str | Dict[Enum, str]]:
        """
        A dictionary containing the parameter descriptions for :meth:`prepare`.

        These should be longer, more detailed descriptions of the parameters of the :meth:`prepare`
        method. The keys must be the names of the parameters, and the values should be strings describing the parameter.

        This method can then be used to generate a help text for the user.

        :return: A dictionary containing the parameter hints for the prepare method.
        """
        return {
            "x10_zip_file": "A zip file containing the VDV data. The zip file must contain all necessary tables for the"
            " VDV 451/452 format. These should be in the form of .x10 files in the top level of the zip"
            " file.",
        }


def validate_zip_file(zipfile: Path) -> bool | Dict[str, str]:
    """
    Validate the zip file.

    :param zipfile: A pathlib.Path object representing the path to the zip file.
    :return: A boolean indicating whether the zip file is valid and either a dictionary containing the error message(s).
    """
    try:
        error_messages = {}
        zip_file = ZipFile(zipfile)
        valid = True
        for entry in zip_file.infolist():
            if (entry.filename.endswith(".x10") or entry.filename.endswith(".X10")) and not entry.is_dir():
                if entry.file_size == 0:
                    valid = False
                    error_messages[entry.filename] = "Empty file"
                elif entry.file_size > 100 * 1024 * 1024:
                    valid = False
                    error_messages[entry.filename] = "File size exceeds 100 MiB"
            else:
                valid = False
                error_messages[entry.filename] = "Invalid file extension or is a directory"
    except Exception as e:
        valid = False
        error_messages["zipfile"] = str(e)

    if not valid:
        return error_messages
    return True


def validate_input_data_vdv_451(
    abs_path_to_folder_with_vdv_files: str | Path,
) -> dict[VDV_Table_Name, VDVTable]:
    """
    Checks if the given directory contains all necessary .x10 files (necessary as in VDV 451/452 specified). Will raise
    an exception if not all necessary tables are present.

    :param abs_path_to_folder_with_vdv_files: The ABSOLUTE path to the directory containing the VDV 451 files
    :return: A dictionary containing the table names as keys and the Table objects as values.
    """
    logger = logging.getLogger(__name__)

    # Create a Pattern to find all .x10 Files in this directory
    # in macOS (unlike windows) searching for *.x10 wiles will NOT find files with the extension .X10.
    # Therefore, we need to search for both, but also filter out the duplicates later as we would otherwise have duplicates in windows
    search_pattern_lowercase = os.path.join(abs_path_to_folder_with_vdv_files, "*.x10")
    search_pattern_uppercase = os.path.join(abs_path_to_folder_with_vdv_files, "*.X10")

    # Find all files that match this pattern.
    x10_files = glob.glob(search_pattern_lowercase) + glob.glob(search_pattern_uppercase)
    x10_files_unique = list(set(x10_files))

    # Iterate through the files, checking whether the neccessary tables are present

    # VDV 451 has two naming schemes. One is the name of the table directly as the name, the other is some number combination,
    # see VDV 451 Chapter 3.1 and 3.2. However, as the name of the table is also included in the file contents, we instead
    # check the contents of each file to determine to which table it belongs.

    all_tables: dict[VDV_Table_Name, VDVTable] = {}
    for abs_file_path in x10_files_unique:
        try:
            eingangsdatentable: VDVTable = check_vdv451_file_header(abs_file_path)

            # Check if the table name is already present in the dictionary (would mean duplicate, two times the same table in the files)
            if eingangsdatentable.table_name in all_tables.keys():
                raise ValueError(
                    "The table " + eingangsdatentable.table_name.value + " is present in multiple files. Aborting."
                )

            else:
                all_tables[eingangsdatentable.table_name] = eingangsdatentable

        except (ValueError, UnicodeDecodeError) as e:
            msg = "While processing " + abs_file_path + " the following exception occurred: "
            logger.debug(msg, exc_info=e)
            continue

    # Required Tables:
    # siehe ganz oben in der File fuer explanation

    if not set(VdvRequiredTables.required_tables.keys()) <= set(all_tables.keys()):
        # Compute all tables that are required but not in the tables in the files, to display them to the user
        missing_tables = set(VdvRequiredTables.required_tables.keys()) - set(all_tables.keys())
        missing_tables_str = " ".join([x.value + ", " for x in missing_tables])
        raise ValueError(
            "Not all necessary tables are present in the directory (or present, but empty). Missing tables are: "
            + missing_tables_str
            + " aborting.",
        )

    # Either REC_FRT_HZT or ORT_HZTF must be present, not both(?)

    if (VDV_Table_Name.REC_FRT_HZT in all_tables.keys()) and (VDV_Table_Name.ORT_HZTF in all_tables.keys()):
        # Both tables present...
        raise ValueError(
            "Either REC_FRT_HZT or ORT_HZTF must be present in the dataset, but both are present. Aborting."
        )

    if (VDV_Table_Name.REC_FRT_HZT not in all_tables.keys()) and (VDV_Table_Name.ORT_HZTF not in all_tables.keys()):
        # Gar keine Haltezeiten dabei
        raise ValueError("Neither REC_FRT_HZT nor ORT_HZTF present in the directory. Aborting.")

    logger.info("All necessary tables are present in the directory.")
    return all_tables


def check_vdv451_file_header(abs_file_path: str) -> VDVTable:
    """
    Checks the contents of a VDV 451 (.x10) file, extracting the table name, character set, column names and data types.
    :param file_path: The ABSOLUTE path to the VDV 451 file
    :return: A VDVTable object containing the table name, character set, column names and data types.

    """

    # 1. Open file and recognize the encoding
    # For VDV 451, either ASCII or ISO8859-1 is allowed as encoding for the table datasets. However, the header is always ASCII (see Ch. 4.1 of VDV 451).
    # Therefore, we open the file with ISO8859-1
    # (and return an error if it is not ASCII or ISO8859-1).
    logger = logging.getLogger(__name__)

    table_name_str = None
    character_set = None
    datatypes = None
    column_names = None

    valid_character_sets = ["ASCII", "ISO8859-1"]

    try:
        with open(abs_file_path, "r", encoding="ISO8859-1") as f:
            for line in f:
                if line.strip().split(";")[0] == "chs":
                    # For these modes, we need to utilize the CSV reader here in order to get rid of the double quote marks enclosing the strings (otherwise, we would have e.g. '"Templin, ZOB"') etc.
                    parts_csvrdr = csv.reader([line], delimiter=";", skipinitialspace=True)
                    parts = list(parts_csvrdr)[0]

                else:
                    # The other modes should be uncritical as they do not contain those double quotes or we dont need the info in them
                    parts = line.strip().split(";")

                # Handling of the line based on the specific "command" (see VDV 451 documentation)
                command = parts[0]

                if command == "tbl":
                    # Get the table name (e.g. 'MENGE_BASIS_VERSIONEN')
                    table_name_str = parts[1].upper().strip()

                elif command == "chs":
                    # Get the character set used in the file
                    character_set = parts[1].upper().strip()

                    # Oftentimes, the Character set is accidentally named as ISO-8859-1 (additional dash).
                    # Fix the character set if it is ISO-8859-1 to the correct form
                    if character_set == "ISO-8859-1":
                        character_set = "ISO8859-1"

                    if character_set not in valid_character_sets:
                        raise ValueError(
                            "The file",
                            abs_file_path,
                            " uses an encoding that is not allowed according to the VDV 451 specification:"
                            + character_set
                            + " does not match 'ASCII' or 'ISO8859-1'.",
                        )

                elif command == "frm":
                    # todo (also for charset) check for double entries of frm, chs, tbl, ..?
                    # Get data formats of the columns (this will be something like ['num[9.0]', 'num[8.0]', 'char[40]', 'num[2.4]'])
                    formats = parts[1:]

                    try:
                        datatypes = parse_datatypes(formats)
                    except ValueError as e:
                        e.add_note(
                            "The file"
                            + str(abs_file_path)
                            + " contains invalid column data types. Please check the formatting of the data types in the file."
                        )
                        raise e

                elif command == "atr":
                    # Get the column names
                    cx = parts[1:]
                    column_names = [x.upper().strip() for x in cx]
                elif command == "rec":
                    if table_name_str is not None and character_set is not None:
                        # We have all necessary information (and it contains at least one record)
                        break

                elif command == "eof":
                    # We reached the end of the file without seeing any records
                    raise ValueError("The file" + str(abs_file_path) + " does not contain any records.")

    except UnicodeDecodeError as e:
        e.add_note(
            "The header of the file"
            + str(abs_file_path)
            + " is using an encoding that contains non-ASCII characters. This is not allowed according to the VDV 451 specification.",
        )
        raise e

    # Raise an error if table name or encoding is not found
    if table_name_str is None:
        msg = f"The file {abs_file_path} does not contain a table name in the header."
        logger.info(msg)
        raise ValueError(msg)
    if character_set is None:
        msg = f"The file {abs_file_path} does not contain a character set in the header."
        logger.info(msg)
        raise ValueError(msg)

    if datatypes is None:
        msg = f"The file {abs_file_path} does not contain the data types of the columns in the header."
        logger.info(msg)
        raise ValueError(msg)

    if table_name_str not in [x.value for x in VDV_Table_Name]:
        raise ValueError(
            "The file" + str(abs_file_path) + " contains an unknown table name: " + table_name_str + " Skipping it."
        )

    if column_names is None:
        raise ValueError(
            "The file"
            + str(abs_file_path)
            + " does not contain the column names in the header. Please check the file and try again."
        )

    if len(column_names) != len(datatypes):
        raise ValueError(
            "The file"
            + str(abs_file_path)
            + " contains an unequal number of column names and column data types in the header: "
            + str(len(column_names))
            + " column names, but "
            + str(len(datatypes))
            + " column data types."
        )

    return VDVTable(
        abs_file_path=abs_file_path,
        character_set=character_set,
        table_name=VDV_Table_Name[table_name_str],
        column_names_and_data_types=list(zip(column_names, datatypes)),
    )


def parse_datatypes(datatype_str: list[str]) -> list[Optional[VDV_Data_Type]]:
    """
    Converts a list of datatype strings in VDV 451 format to a list of Python/Numpy datatypes
    e.g., turn something like ['num[9.0]', 'char[40]', 'num[2.0]'] into ['int', 'string', 'int']

    (We do this as we will later convert the column datatypes to the correct Python/Numpy datatypes)
    So for every column in the VDV 451 file, check if 'num', 'int' or 'float'.

    :param datatype_str: a list with the datatypes from the VDV 451 file, (with each datatype as a string, e.g. 'char[40]')
    :return: a list of python datatypes, but as strings
    """

    # todo add logger?
    logger = logging.getLogger(__name__)

    dtypes: list[Optional[VDV_Data_Type]] = []
    for part in datatype_str:
        part = part.lstrip()  # remove leading spaces

        # check if the datatype is valid (e.g. 'num[9.0]' or 'char[40]' etc.)
        # according to the VDV 451 specification, only 'char[n]' and 'num[n.0]' are allowed

        regex = r"(char\[[0-9]+\]|num\[[0-9]+.[0-9]+\])+"

        if not re.match(regex, part):
            # Avoid the program to crash if the datatype is invalid, but still log a warning
            # Sometimes, there are floats used for additional columns (columns not formally included the VDV 452 specification)
            dtypes.append(None)
            # todo genauere angabe, in welcher Datei / Spalte es auftrat?
            msg = f"Invalid datatype formatting in VDV 451 file: {part} does not match 'char[n]' or 'num[n.n]'. Column will not be imported."
            logger.warning(msg)
            continue

        regex_for_char = r"char\[[0-9]+\]"
        regex_for_int = r"num\[[0-9]+.0\]"
        regex_for_float = r"num\[[0-9]+.[0-9]+\]"

        if re.match(regex_for_char, part):
            dtypes.append(VDV_Data_Type.CHAR)
        elif re.match(regex_for_int, part):
            dtypes.append(VDV_Data_Type.INT)
        elif re.match(regex_for_float, part):
            dtypes.append(VDV_Data_Type.FLOAT)

    return dtypes


def import_vdv452_table_records(EingangsdatenTabelle: VDVTable) -> list[VdvBaseObject]:
    """
    Imports the records of a VDV 451 table into the database.
    :param EingangsdatenTabelle: The EingangsdatenTabelle object containing the table name and the path to the file
    :return: None
    """
    logger = logging.getLogger(__name__)

    # Open the file
    with open(EingangsdatenTabelle.abs_file_path, "r", encoding=EingangsdatenTabelle.character_set) as f:
        reader = csv.reader(f, delimiter=";", skipinitialspace=True)
        dict_list = []
        for row in reader:
            if len(row) == 0 or row[0].strip() != "rec":
                logger.debug("Skipping line: " + str(row))
                continue

            # Remove the 'rec' from the row
            row_data = row[1:]

            # create the json obj and give every column value the correct datatype
            e_data: Dict[str, str | int | float | None] = {}

            if len(row_data) != len(EingangsdatenTabelle.column_names_and_data_types):
                raise ValueError(
                    "The file"
                    + str(EingangsdatenTabelle.abs_file_path)
                    + " contains an record that has more or less columns than the header specifies. "
                    + "The record contains "
                    + str(row_data)
                    + ", aborting."
                )

            for i_col in range(0, len(row_data)):
                column_name = EingangsdatenTabelle.column_names_and_data_types[i_col][0]
                column_data_type = EingangsdatenTabelle.column_names_and_data_types[i_col][1]

                if row_data[i_col].strip() == "":
                    # Everything that has "no" value in the VDV 451 file is turned into a None
                    # NULL Entry (Also possible for numbers - thats why it is done BEFORE the Int conversion!)
                    e_data[column_name] = None

                elif column_data_type is None:
                    # Skip the column as it has an invalid data type.
                    continue

                elif column_data_type == VDV_Data_Type.INT:
                    try:
                        e_data[column_name] = int(row_data[i_col])
                    except ValueError as e:
                        e.add_note(
                            "The file"
                            + str(EingangsdatenTabelle.abs_file_path)
                            + " contains a non-numeric value in a column that is specified as numeric. Aborting."
                        )
                        raise e
                elif column_data_type == VDV_Data_Type.FLOAT:
                    try:
                        e_data[column_name] = float(row_data[i_col])
                    except ValueError as e:
                        e.add_note(
                            "The file"
                            + str(EingangsdatenTabelle.abs_file_path)
                            + " contains a non-numeric value in a column that is specified as numeric. Aborting."
                        )
                        raise e
                elif column_data_type == VDV_Data_Type.CHAR:
                    e_data[column_name] = row_data[i_col]
                else:
                    raise ValueError(
                        "The file"
                        + str(EingangsdatenTabelle.abs_file_path)
                        + " contains a column with an invalid data type: "
                        + str(column_data_type)
                        + ". Aborting."
                    )
            dict_list.append(e_data)

        # Now that we have created a nice dictionary, turn it into an object of the corresponding dataclass
        match EingangsdatenTabelle.table_name:
            case VDV_Table_Name.BASIS_VER_GUELTIGKEIT:
                # At the current time, we only support one distinct entry for this table.
                # If there are multiple. raise an error.
                # However, if the same entry is present multiple times, we do not raise an error.
                # So we need to turn the list of dictionaries inso a set of objects and check if the length is 1.
                objects: List[VdvBaseObject] = []
                for d in dict_list:
                    objects.append(BasisVerGueltigkeit.from_dict(d))
                if len(set(objects)) != 1:
                    raise ValueError(
                        "The table"
                        + str(EingangsdatenTabelle.table_name)
                        + " contains multiple distinct entries. Only one entry is allowed. Aborting."
                    )
                return objects

            case VDV_Table_Name.FIRMENKALENDER:
                return [Firmenkalender.from_dict(d) for d in dict_list]
            case VDV_Table_Name.REC_ORT:
                return [RecOrt.from_dict(d) for d in dict_list]
            case VDV_Table_Name.MENGE_FZG_TYP:
                return [MengeFzgTyp.from_dict(d) for d in dict_list]
            case VDV_Table_Name.REC_SEL:
                return [RecSel.from_dict(d) for d in dict_list]
            case VDV_Table_Name.SEL_FZT_FELD:
                return [SelFztFeld.from_dict(d) for d in dict_list]
            case VDV_Table_Name.LID_VERLAUF:
                return [LidVerlauf.from_dict(d) for d in dict_list]
            case VDV_Table_Name.REC_FRT:
                return [RecFrt.from_dict(d) for d in dict_list]
            case VDV_Table_Name.REC_UMLAUF:
                return [RecUmlauf.from_dict(d) for d in dict_list]
            case VDV_Table_Name.REC_LID:
                return [RecLid.from_dict(d) for d in dict_list]
            case VDV_Table_Name.REC_FRT_HZT:
                return [RecFrtHzt.from_dict(d) for d in dict_list]
            case VDV_Table_Name.ORT_HZTF:
                return [OrtHztf.from_dict(d) for d in dict_list]
            case _:  # default case
                return []
